from octue import exceptions
from octue.mixins.filterable import Filterable
from octue.resources.tag import TagSet
from tests.base import BaseTestCase


class FilterableSubclass(Filterable):
    def __init__(self, name=None, is_alive=None, iterable=None, age=None, owner=None):
        self.name = name
        self.is_alive = is_alive
        self.iterable = iterable
        self.age = age
        self.owner = owner


class TestFilterable(BaseTestCase):
    def test_error_raised_when_invalid_filter_name_received(self):
        """ Ensure an error is raised when an invalid filter name is provided. """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass().satisfies(filter_name="invalid_filter_name", filter_value=None)

    def test_error_raised_when_non_existent_attribute_name_received(self):
        """ Ensure an error is raised when a non-existent attribute name is used in the filter name. """
        with self.assertRaises(AttributeError):
            FilterableSubclass().satisfies(filter_name="boogaloo__is_a_dance", filter_value=True)

    def test_error_raised_when_valid_but_non_existent_filter_name_received(self):
        """ Ensure an error is raised when a valid but non-existent filter name is received. """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass().satisfies(filter_name="age__is_secret", filter_value=True)

    def test_error_raised_when_attribute_type_has_no_filters_defined(self):
        """Ensure an error is raised when a filter for an attribute whose type doesn't have any filters defined is
        received.
        """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass(age=lambda: None).satisfies(filter_name="age__equals", filter_value=True)

    def test_bool_filters(self):
        """ Test that the boolean filters work as expected. """
        filterable_thing = FilterableSubclass(is_alive=True)
        self.assertTrue(filterable_thing.satisfies("is_alive__is", True))
        self.assertFalse(filterable_thing.satisfies("is_alive__is", False))
        self.assertTrue(filterable_thing.satisfies("is_alive__is_not", False))
        self.assertFalse(filterable_thing.satisfies("is_alive__is_not", True))

    def test_str_filters(self):
        """ Test that the string filters work as expected. """
        filterable_thing = FilterableSubclass(name="Michael")
        self.assertTrue(filterable_thing.satisfies("name__icontains", "m"))
        self.assertFalse(filterable_thing.satisfies("name__icontains", "d"))
        self.assertTrue(filterable_thing.satisfies("name__not_icontains", "d"))
        self.assertFalse(filterable_thing.satisfies("name__not_icontains", "m"))
        self.assertTrue(filterable_thing.satisfies("name__contains", "M"))
        self.assertFalse(filterable_thing.satisfies("name__contains", "d"))
        self.assertTrue(filterable_thing.satisfies("name__ends_with", "l"))
        self.assertFalse(filterable_thing.satisfies("name__ends_with", "M"))
        self.assertTrue(filterable_thing.satisfies("name__not_ends_with", "M"))
        self.assertFalse(filterable_thing.satisfies("name__not_ends_with", "l"))
        self.assertTrue(filterable_thing.satisfies("name__starts_with", "M"))
        self.assertFalse(filterable_thing.satisfies("name__starts_with", "l"))
        self.assertTrue(filterable_thing.satisfies("name__not_starts_with", "l"))
        self.assertFalse(filterable_thing.satisfies("name__not_starts_with", "M"))
        self.assertTrue(filterable_thing.satisfies("name__equals", "Michael"))
        self.assertFalse(filterable_thing.satisfies("name__equals", "Clive"))
        self.assertTrue(filterable_thing.satisfies("name__not_equals", "Clive"))
        self.assertFalse(filterable_thing.satisfies("name__not_equals", "Michael"))
        self.assertTrue(filterable_thing.satisfies("name__iequals", "michael"))
        self.assertFalse(filterable_thing.satisfies("name__iequals", "James"))
        self.assertTrue(filterable_thing.satisfies("name__not_iequals", "James"))
        self.assertFalse(filterable_thing.satisfies("name__not_iequals", "michael"))
        self.assertTrue(filterable_thing.satisfies("name__is", "Michael"))
        self.assertFalse(filterable_thing.satisfies("name__is", "Clive"))
        self.assertTrue(filterable_thing.satisfies("name__is_not", "Clive"))
        self.assertFalse(filterable_thing.satisfies("name__is_not", "Michael"))
        self.assertTrue(filterable_thing.satisfies("name__lt", "Noel"))
        self.assertFalse(filterable_thing.satisfies("name__lt", "Harry"))
        self.assertTrue(filterable_thing.satisfies("name__lte", "Michael"))
        self.assertFalse(filterable_thing.satisfies("name__lte", "Harry"))
        self.assertTrue(filterable_thing.satisfies("name__gt", "Clive"))
        self.assertFalse(filterable_thing.satisfies("name__gt", "Noel"))
        self.assertTrue(filterable_thing.satisfies("name__gte", "Michael"))
        self.assertFalse(filterable_thing.satisfies("name__gte", "Noel"))

    def test_none_filters(self):
        """ Test that the None filters work as expected. """
        filterable_thing = FilterableSubclass(owner=None)
        self.assertTrue(filterable_thing.satisfies("owner__is", None))
        self.assertFalse(filterable_thing.satisfies("owner__is", True))
        self.assertTrue(filterable_thing.satisfies("owner__is_not", True))
        self.assertFalse(filterable_thing.satisfies("owner__is_not", None))

    def test_number_filters_with_integers_and_floats(self):
        """ Test that the number filters work as expected for integers and floats. """
        for age in (5, 5.2):
            filterable_thing = FilterableSubclass(age=age)
            self.assertTrue(filterable_thing.satisfies("age__equals", age))
            self.assertFalse(filterable_thing.satisfies("age__equals", 63))
            self.assertTrue(filterable_thing.satisfies("age__not_equals", 63))
            self.assertFalse(filterable_thing.satisfies("age__not_equals", age))
            self.assertTrue(filterable_thing.satisfies("age__lt", 6))
            self.assertFalse(filterable_thing.satisfies("age__lt", 0))
            self.assertTrue(filterable_thing.satisfies("age__lte", age))
            self.assertFalse(filterable_thing.satisfies("age__lte", 0))
            self.assertTrue(filterable_thing.satisfies("age__gt", 4))
            self.assertFalse(filterable_thing.satisfies("age__gt", 63))
            self.assertTrue(filterable_thing.satisfies("age__gte", age))
            self.assertFalse(filterable_thing.satisfies("age__gte", 63))
            self.assertTrue(filterable_thing.satisfies("age__is", age))
            self.assertFalse(filterable_thing.satisfies("age__is", 63))
            self.assertTrue(filterable_thing.satisfies("age__is_not", 63))
            self.assertFalse(filterable_thing.satisfies("age__is_not", age))

    def test_iterable_filters(self):
        """ Test that the iterable filters work as expected with lists, sets, and tuples. """
        for iterable in ([1, 2, 3], {1, 2, 3}, (1, 2, 3)):
            filterable_thing = FilterableSubclass(iterable=iterable)
            self.assertTrue(filterable_thing.satisfies("iterable__contains", 1))
            self.assertFalse(filterable_thing.satisfies("iterable__contains", 5))
            self.assertTrue(filterable_thing.satisfies("iterable__not_contains", 5))
            self.assertFalse(filterable_thing.satisfies("iterable__not_contains", 1))
            self.assertTrue(filterable_thing.satisfies("iterable__is", iterable))
            self.assertFalse(filterable_thing.satisfies("iterable__is", None))
            self.assertTrue(filterable_thing.satisfies("iterable__is_not", None))
            self.assertFalse(filterable_thing.satisfies("iterable__is_not", iterable))

    def test_tag_set_filters(self):
        """ Test the filters for TagSet. """
        filterable_thing = FilterableSubclass(iterable=TagSet({"fred", "charlie"}))
        self.assertTrue(filterable_thing.satisfies("iterable__any_tag_contains", "a"))
        self.assertFalse(filterable_thing.satisfies("iterable__any_tag_contains", "z"))
        self.assertTrue(filterable_thing.satisfies("iterable__not_any_tag_contains", "z"))
        self.assertFalse(filterable_thing.satisfies("iterable__not_any_tag_contains", "a"))
        self.assertTrue(filterable_thing.satisfies("iterable__any_tag_starts_with", "f"))
        self.assertFalse(filterable_thing.satisfies("iterable__any_tag_starts_with", "e"))
        self.assertTrue(filterable_thing.satisfies("iterable__any_tag_ends_with", "e"))
        self.assertFalse(filterable_thing.satisfies("iterable__any_tag_ends_with", "i"))
        self.assertTrue(filterable_thing.satisfies("iterable__not_any_tag_starts_with", "e"))
        self.assertFalse(filterable_thing.satisfies("iterable__not_any_tag_starts_with", "f"))
        self.assertTrue(filterable_thing.satisfies("iterable__not_any_tag_ends_with", "i"))
        self.assertFalse(filterable_thing.satisfies("iterable__not_any_tag_ends_with", "e"))

    def test_filtering_different_attributes_on_same_instance(self):
        """ Ensure all filterable attributes on an instance can be checked for filter satisfaction. """
        filterable_thing = FilterableSubclass(name="Fred", is_alive=True, iterable={1, 2, 3}, age=5.2, owner=None)
        self.assertTrue(filterable_thing.satisfies("name__icontains", "f"))
        self.assertTrue(filterable_thing.satisfies("name__not_icontains", "j"))
        self.assertFalse(filterable_thing.satisfies("is_alive__is", False))
        self.assertTrue(filterable_thing.satisfies("iterable__contains", 3))
        self.assertTrue(filterable_thing.satisfies("age__equals", 5.2))
        self.assertTrue(filterable_thing.satisfies("age__not_equals", 5))
        self.assertTrue(filterable_thing.satisfies("owner__is", None))



import json
import logging
import os
from tempfile import TemporaryDirectory

from octue.mixins import Serialisable
from ..base import BaseTestCase


class Inherit(Serialisable):
    def __init__(self):
        super().__init__()
        self.id = "id"
        self.logger = logging.getLogger("test_returns_primitive_without_logger_or_protected_fields")
        self.field_to_serialise = 0
        self._field_not_to_serialise = 1


class InheritWithFieldsToSerialise(Inherit):
    _SERIALISE_FIELDS = ("field_to_serialise",)


class SerialisableTestCase(BaseTestCase):
    def test_instantiates_with_no_args(self):
        """Ensures the class instantiates without arguments"""
        Serialisable()

    def test_raises_attribute_error_with_missing_logger(self):
        """Ensures class instantiates with a string uuid"""
        resource = Serialisable()
        with self.assertRaises(AttributeError) as error:
            resource.serialise()

        self.assertIn("'Serialisable' object has no attribute 'logger'", error.exception.args[0])

    def test_returns_primitive_without_logger_or_protected_fields(self):
        """Ensures class instantiates with a UUID()"""
        resource = Inherit()
        serialised = resource.serialise()
        self.assertTrue("id" in serialised.keys())
        self.assertTrue("field_to_serialise" in serialised.keys())
        self.assertFalse("logger" in serialised.keys())
        self.assertFalse("_field_not_to_serialise" in serialised.keys())

    def test_serialise_only_attrs(self):
        """Restricts the id field, which would normally be serialised"""
        resource = InheritWithFieldsToSerialise()
        serialised = resource.serialise()
        self.assertFalse("id" in serialised.keys())
        self.assertTrue("field_to_serialise" in serialised.keys())
        self.assertFalse("logger" in serialised.keys())
        self.assertFalse("_field_not_to_serialise" in serialised.keys())

    def test_serialise_to_string(self):
        """Restricts the id field, which would normally be serialised"""
        resource = InheritWithFieldsToSerialise()
        serialised = resource.serialise(to_string=True)
        self.assertIsInstance(serialised, str)

    def test_serialise_to_file(self):
        """Restricts the id field, which would normally be serialised"""

        with TemporaryDirectory() as dir_name:
            file_name = os.path.join(dir_name, "test_serialise_to_file.json")
            resource = Inherit()
            resource.to_file(file_name)

            # Test that the file was written with appropriate contents
            with open(file_name, "r") as fp:
                serialised = json.load(fp)

            self.assertTrue("id" in serialised.keys())
            self.assertTrue("field_to_serialise" in serialised.keys())
            self.assertFalse("logger" in serialised.keys())
            self.assertFalse("_field_not_to_serialise" in serialised.keys())



import json
import logging
import os
import warnings

from octue import definitions
from octue.exceptions import BrokenSequenceException, InvalidInputException, UnexpectedNumberOfResultsException
from octue.mixins import Hashable, Identifiable, Loggable, Pathable, Serialisable, Taggable
from octue.resources.datafile import Datafile
from octue.resources.filter_containers import FilterSet
from octue.utils.cloud import storage
from octue.utils.cloud.storage.client import GoogleCloudStorageClient


module_logger = logging.getLogger(__name__)


DATAFILES_DIRECTORY = "datafiles"


class Dataset(Taggable, Serialisable, Pathable, Loggable, Identifiable, Hashable):
    """A representation of a dataset, containing files, tags, etc

    This is used to read a list of files (and their associated properties) into octue analysis, or to compile a
    list of output files (results) and their properties that will be sent back to the octue system.
    """

    _FILTERSET_ATTRIBUTE = "files"
    _ATTRIBUTES_TO_HASH = "files", "name", "tags"

    def __init__(self, name=None, id=None, logger=None, path=None, path_from=None, tags=None, **kwargs):
        """Construct a Dataset"""
        super().__init__(id=id, logger=logger, tags=tags, path=path, path_from=path_from)

        self._name = name

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then remove this hackjob.
        self.files = FilterSet()

        for file in kwargs.pop("files", list()):
            if isinstance(file, Datafile):
                self.files.add(file)
            else:
                self.files.add(Datafile(**file, path_from=self))

        self.__dict__.update(**kwargs)

    @classmethod
    def from_cloud(cls, project_name, bucket_name, path_to_dataset_directory):
        """Instantiate a Dataset from Google Cloud storage.

        :param str project_name:
        :param str bucket_name:
        :param str path_to_dataset_directory: path to dataset directory (directory containing dataset's files)
        :return Dataset:
        """
        storage_client = GoogleCloudStorageClient(project_name=project_name)

        serialised_dataset = json.loads(
            storage_client.download_as_string(
                bucket_name=bucket_name,
                path_in_bucket=storage.path.join(path_to_dataset_directory, definitions.DATASET_FILENAME),
            )
        )

        datafiles = FilterSet()

        for file in serialised_dataset["files"]:
            file_bucket_name, path = storage.path.split_bucket_name_from_gs_path(file)

            datafiles.add(
                Datafile.from_cloud(project_name=project_name, bucket_name=file_bucket_name, datafile_path=path)
            )

        return Dataset(
            id=serialised_dataset["id"],
            name=serialised_dataset["name"],
            hash_value=serialised_dataset["hash_value"],
            path=storage.path.generate_gs_path(bucket_name, path_to_dataset_directory),
            tags=json.loads(serialised_dataset["tags"]),
            files=datafiles,
        )

    def to_cloud(self, project_name, bucket_name, output_directory):
        """Upload a dataset to a cloud location.

        :param str project_name:
        :param str bucket_name:
        :param str output_directory:
        :return str: gs:// path for dataset
        """
        files = []

        for datafile in self.files:
            datafile_path = datafile.to_cloud(
                project_name, bucket_name, path_in_bucket=storage.path.join(output_directory, self.name, datafile.name)
            )

            files.append(datafile_path)

        serialised_dataset = self.serialise()
        serialised_dataset["files"] = sorted(files)
        del serialised_dataset["absolute_path"]
        del serialised_dataset["path"]

        GoogleCloudStorageClient(project_name=project_name).upload_from_string(
            string=json.dumps(serialised_dataset),
            bucket_name=bucket_name,
            path_in_bucket=storage.path.join(output_directory, self.name, definitions.DATASET_FILENAME),
        )

        return storage.path.generate_gs_path(bucket_name, output_directory, self.name)

    @property
    def name(self):
        return self._name or os.path.split(os.path.abspath(os.path.split(self.path)[-1]))[-1]

    def __iter__(self):
        yield from self.files

    def __len__(self):
        return len(self.files)

    def add(self, *args, **kwargs):
        """Add a data/results file to the manifest

        Usage:
            my_file = octue.DataFile(...)
            my_manifest.add(my_file)

            # or more simply
            my_manifest.add(**{...}) which implicitly creates the datafile from the starred list of input arguments

        """
        if len(args) > 1:
            # Recurse to allow addition of many files at once
            for arg in args:
                self.add(arg, **kwargs)
        elif len(args) > 0:
            if not isinstance(args[0], Datafile):
                raise InvalidInputException(
                    'Object "{}" must be of class Datafile to add it to a Dataset'.format(args[0])
                )
            self.files.add(args[0])

        else:
            # Add a single file, constructed by passing the arguments through to DataFile()
            self.files.add(Datafile(**kwargs))

    def append(self, *args, **kwargs):
        warnings.warn(
            "The `Dataset.append` method has been deprecated and replaced with `Dataset.add` to reflect that Datafiles "
            "are stored in a set and not a list. Calls to `Dataset.append` will be redirected to the new method for "
            "now, but please use `Datafile.add` in future.",
            DeprecationWarning,
        )
        self.files.add(*args, **kwargs)

    def get_files(self, field_lookup, filter_value=None):
        warnings.warn(
            "The `Dataset.get_files` method has been deprecated and replaced with `Dataset.files.filter`, which has "
            "the same interface but with the `field_lookup` argument renamed to `filter_name`. Calls to "
            "`Dataset.get_files` will be redirected to the new method for now, but please use `Datafile.files.filter` "
            "in future.",
            DeprecationWarning,
        )
        return self.files.filter(filter_name=field_lookup, filter_value=filter_value)

    def get_file_sequence(self, filter_name, filter_value=None, strict=True):
        """Get an ordered sequence of files matching a criterion

        Accepts the same search arguments as `get_files`.

        :parameter strict: If True, applies a check that the resulting file sequence begins at 0 and ascends uniformly
        by 1
        :type strict: bool

        :returns: Sorted list of Datafiles
        :rtype: list(Datafile)
        """

        results = self.files.filter(filter_name=filter_name, filter_value=filter_value)
        results = results.filter("sequence__is_not", None)

        def get_sequence_number(file):
            return file.sequence

        # Sort the results on ascending sequence number
        results = sorted(results, key=get_sequence_number)

        # Check sequence is unique and sequential
        if strict:
            index = -1
            for result in results:
                index += 1
                if result.sequence != index:
                    raise BrokenSequenceException("Filtered file sequence numbers do not monotonically increase from 0")

        return results

    def get_file_by_tag(self, tag_string):
        """Gets a data file from a manifest by searching for files with the provided tag(s)

        Gets exclusively one file; if no file or more than one file is found this results in an error.

        :param tag_string: if this string appears as an exact match in the tags
        :return: DataFile object
        """
        results = self.files.filter(filter_name="tags__contains", filter_value=tag_string)
        if len(results) > 1:
            raise UnexpectedNumberOfResultsException("More than one result found when searching for a file by tag")
        elif len(results) == 0:
            raise UnexpectedNumberOfResultsException("No files found with this tag")

        return results.pop()



from octue import exceptions
from octue.mixins import MixinBase, Taggable
from octue.resources.tag import Tag, TagSet
from ..base import BaseTestCase


class MyTaggable(Taggable, MixinBase):
    pass


class TaggableTestCase(BaseTestCase):
    def test_instantiates(self):
        """Ensures the class instantiates without arguments"""
        Taggable()

    def test_instantiates_with_tags(self):
        """Ensures datafile inherits correctly from the Taggable class and passes arguments through"""
        taggable = MyTaggable(tags="")
        self.assertEqual(len(taggable.tags), 0)

        taggable = MyTaggable(tags=None)
        self.assertEqual(len(taggable.tags), 0)

        taggable = MyTaggable(tags="a b c")
        self.assertEqual(set(taggable.tags), {Tag("a"), Tag("b"), Tag("c")})

        with self.assertRaises(exceptions.InvalidTagException):
            MyTaggable(tags=":a b c")

    def test_instantiates_with_tag_set(self):
        """Ensures datafile inherits correctly from the Taggable class and passes arguments through"""
        taggable_1 = MyTaggable(tags="")
        self.assertIsInstance(taggable_1.tags, TagSet)
        taggable_2 = MyTaggable(tags=taggable_1.tags)
        self.assertFalse(taggable_1 is taggable_2)

    def test_fails_to_instantiates_with_non_iterable(self):
        """Ensures datafile inherits correctly from the Taggable class and passes arguments through"""

        class NoIter:
            pass

        with self.assertRaises(exceptions.InvalidTagException) as error:
            MyTaggable(tags=NoIter())

        self.assertIn(
            "Tags must be expressed as a whitespace-delimited string or an iterable of strings", error.exception.args[0]
        )

    def test_reset_tags(self):
        """Ensures datafile inherits correctly from the Taggable class and passes arguments through"""
        taggable = MyTaggable(tags="a b")
        taggable.tags = "b c"
        self.assertEqual(set(taggable.tags), {Tag("b"), Tag("c")})

    def test_valid_tags(self):
        """Ensures valid tags do not raise an error"""
        taggable = MyTaggable()
        taggable.add_tags("a-valid-tag")
        taggable.add_tags("a:tag")
        taggable.add_tags("a:-tag")  # <--- yes, this is valid deliberately as it allows people to do negation
        taggable.add_tags("a1829tag")
        taggable.add_tags("1829")
        taggable.add_tags("number:1829")
        taggable.add_tags("multiple:discriminators:used")
        self.assertEqual(
            set(taggable.tags),
            {
                Tag("a-valid-tag"),
                Tag("a:tag"),
                Tag("a:-tag"),
                Tag("a1829tag"),
                Tag("1829"),
                Tag("number:1829"),
                Tag("multiple:discriminators:used"),
            },
        )

    def test_invalid_tags(self):
        """Ensures invalid tags raise an error"""
        taggable = MyTaggable()

        for tag in "-bah", "humbug:", "SHOUTY", r"back\slashy", {"not-a": "string"}:
            with self.assertRaises(exceptions.InvalidTagException):
                taggable.add_tags(tag)

    def test_mixture_valid_invalid(self):
        """Ensures that adding a variety of tags, some of which are invalid, doesn't partially add them to the object"""
        taggable = MyTaggable()
        taggable.add_tags("first-valid-should-be-added")
        try:
            taggable.add_tags("second-valid-should-not-be-added-because", "-the-third-is-invalid:")

        except exceptions.InvalidTagException:
            pass

        self.assertEqual({Tag("first-valid-should-be-added")}, set(taggable.tags))



from octue import exceptions


def _filter(instance, filter_name=None, filter_value=None):
    """Returns a new instance containing only the Filterables to which the given filter criteria apply.

    Say we want to filter by files whose extension equals "csv". We want to be able to do...

    ds = DataSet(... initialise with csv files and other files)
    Example of chaining:

    ds.filter('files__extension_equals', 'csv')
    is equvalent to
    ds.files.filter(extension__equals', 'csv')
    >>> FilterSet containing a set of datafiles


    ds.filter('files__extension_equals', 'csv')
    is equvalent to
    ds.files.filter(extension__equals', 'csv')
    """
    return instance.__class__((item for item in instance if item.satisfies(filter_name, filter_value)))


def _order_by(instance, attribute_name, reverse=False):
    """Order the instance by the given attribute_name, returning the instance's elements as a new FilterList (not a
    FilterSet.
    """
    try:
        return FilterList(sorted(instance, key=lambda item: getattr(item, attribute_name), reverse=reverse))
    except AttributeError:
        raise exceptions.InvalidInputException(
            f"An attribute named {attribute_name!r} does not exist on one or more members of {instance!r}."
        )


class FilterSet(set):
    filter = _filter
    order_by = _order_by


class FilterList(list):
    filter = _filter
    order_by = _order_by



import importlib
import logging
import os
import sys

from octue.logging_handlers import apply_log_handler
from octue.resources import Child
from octue.resources.analysis import CLASS_MAP, Analysis
from octue.utils import gen_uuid
from twined import Twine


package_logger = logging.getLogger("octue")
module_logger = logging.getLogger(__name__)


class Runner:
    """Runs analyses in the app framework

    The Runner class provides a set of configuration parameters for use by your application, together with a range of
    methods for managing input and output file parsing as well as controlling logging.

    :parameter twine: string path to the twine file, or a string containing valid twine json

    :parameter paths: string or dict. If a string, contains a single path to an existing data directory where
    (if not already present), subdirectories 'configuration', 'input', 'tmp', 'log' and 'output' will be created. If a
    dict, it should contain all of those keys, with each of their values being a path to a directory (which will be
    recursively created if it doesn't exist)

    :parameter configuration_values: The strand data. Can be expressed as a string path of a *.json file (relative
    or absolute), as an open file-like object (containing json data), as a string of json data or as an
    already-parsed dict.

    :parameter configuration_manifest: The strand data. Can be expressed as a string path of a *.json file
    (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
    already-parsed dict.

    :parameter skip_file_checks: If true, skip the check that all files in the manifest are present on disc - this
    can be an extremely long process for large datasets.
    """

    def __init__(
        self,
        app_src,
        twine="twine.json",
        configuration_values=None,
        configuration_manifest=None,
        output_manifest_path=None,
        credentials=None,
        children=None,
        skip_checks=False,
        log_level=logging.INFO,
        handler=None,
        show_twined_logs=False,
    ):
        """ Constructor for the Runner class. """
        self.app_src = app_src
        self.output_manifest_path = output_manifest_path
        self.credentials = credentials
        self.children = children
        self.skip_checks = skip_checks

        # Store the log level (same log level used for all analyses)
        self._log_level = log_level
        self.handler = handler

        if show_twined_logs:
            apply_log_handler(logger=package_logger, handler=self.handler)

        # Ensure the twine is present and instantiate it
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        package_logger.debug("Parsed twine with strands %r", self.twine.available_strands)

        # Validate and initialise configuration data
        self.configuration = self.twine.validate(
            configuration_values=configuration_values,
            configuration_manifest=configuration_manifest,
            cls=CLASS_MAP,
        )
        package_logger.debug("Configuration validated.")

        # Set path for configuration manifest.
        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        self.configuration["configuration_manifest"] = self._update_manifest_path(
            self.configuration.get("configuration_manifest", None),
            configuration_manifest,
        )

        # Store the log level (same log level used for all analyses)
        self._log_level = log_level
        self.handler = handler

        if show_twined_logs:
            apply_log_handler(logger=package_logger, handler=self.handler, log_level=self._log_level)
            package_logger.info(
                "Showing package logs as well as analysis logs (the package logs are recommended for software "
                "engineers but may still be useful to app development by scientists."
            )

    @staticmethod
    def _update_manifest_path(manifest, pathname):
        """A Quick hack to stitch the new Pathable functionality in the 0.1.4 release into the CLI and runner.

        The way we define a manifest path can be more robustly implemented as we migrate functionality into the twined
        library

        :param manifest:
        :type manifest:
        :param pathname:
        :type pathname:
        :return:
        :rtype:
        """
        if manifest is not None and hasattr(pathname, "endswith"):
            if pathname.endswith(".json"):
                manifest.path = os.path.split(pathname)[0]

        # Otherwise do nothing and rely on manifest having its path variable set already
        return manifest

    def run(self, analysis_id=None, input_values=None, input_manifest=None):
        """Run an analysis

        :parameter app_src: Either: an instance of the AppFrom manager class which has a run() method, or
        a function which accepts a single parameter (the instantiated analysis), or a string pointing
        to an application folder (which should contain an 'app.py' function like the templates).
        :type app_src: Union[AppFrom, function, str]

        :parameter input_values: The input_values strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type input_values: Union[str, dict]

        :parameter input_manifest: The input_manifest strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type input_manifest: Union[str, Manifest]

        :parameter credentials: The credentials strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type credentials: Union[str, dict]

        :parameter children: The children strand data. Can be expressed as a string path of a *.json file
        (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an
        already-parsed dict.
        :type children: Union[str, dict]

        :parameter output_manifest_path: Path where output data will be written
        :type output_manifest_path: Union[str, path-like]

        :parameter handler: the logging.Handler instance which will be used to handle logs for this analysis run.
        handlers can be created as per the logging cookbook https://docs.python.org/3/howto/logging-cookbook.html but
        should use the format defined above in LOG_FORMAT.
        :type handler: logging.Handler

        :return: None
        """
        inputs = self.twine.validate(
            input_values=input_values,
            input_manifest=input_manifest,
            credentials=self.credentials,
            children=self.children,
            cls=CLASS_MAP,
            allow_missing=False,
            allow_extra=False,
        )
        package_logger.debug("Inputs validated.")

        if inputs["children"] is not None:
            inputs["children"] = {
                child["key"]: Child(name=child["key"], id=child["id"], backend=child["backend"])
                for child in inputs["children"]
            }

        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        inputs["input_manifest"] = self._update_manifest_path(
            inputs.get("input_manifest", None),
            input_manifest,
        )

        outputs_and_monitors = self.twine.prepare("monitors", "output_values", "output_manifest", cls=CLASS_MAP)

        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        outputs_and_monitors["output_manifest"] = self._update_manifest_path(
            outputs_and_monitors.get("output_manifest", None),
            self.output_manifest_path,
        )

        analysis_id = str(analysis_id) if analysis_id else gen_uuid()
        analysis_logger = logging.getLogger(f"analysis-{analysis_id}")
        apply_log_handler(logger=analysis_logger, handler=self.handler, log_level=self._log_level)

        analysis = Analysis(
            id=analysis_id,
            logger=analysis_logger,
            twine=self.twine,
            skip_checks=self.skip_checks,
            **self.configuration,
            **inputs,
            **outputs_and_monitors,
        )

        try:
            if hasattr(self.app_src, "run"):
                self.app_src.run(analysis)
            elif isinstance(self.app_src, str):
                with AppFrom(self.app_src) as app:
                    app.run(analysis)
            else:
                self.app_src(analysis)

        except Exception as e:
            analysis_logger.error(str(e))
            raise e

        return analysis


def unwrap(fcn):
    """Recurse through wrapping to get the raw function without decorators."""
    if hasattr(fcn, "__wrapped__"):
        return unwrap(fcn.__wrapped__)
    return fcn


class AppFrom:
    """Context manager that imports module 'app' from user's code base at a location app_path.

     The manager will issue a warning if an existing module called "app" is already loaded.

     The manager makes a temporary addition to the system path (to ensure app is loaded from the correct path)

     The manager will unload the module (by deleting it from sys.modules) on exit, enabling

    with AppFrom('/path/to/dir') as app:
        Runner().run(app)

    """

    def __init__(self, app_path="."):
        self.app_path = os.path.abspath(os.path.normpath(app_path))
        module_logger.debug("Initialising AppFrom context at app_path %s", self.app_path)
        self.app_module = None

    def __enter__(self):
        # Warn on an app present on the system path
        if "app" in sys.modules.keys():
            module_logger.warning(
                "Module 'app' already on system path. Using 'AppFrom' context will yield unexpected results. Avoid using 'app' as a python module, except for your main entrypoint"
            )

        # Insert the present directory first on the system path
        sys.path.insert(0, self.app_path)

        # Import the app from the present directory
        self.app_module = importlib.import_module("app")

        # Immediately clean up the entry to the system path (don't use "remove" because if the user has it in their
        # path, this'll be an unexpected side effect, and don't do it in cleanup in case the called code inserts a path)
        sys.path.pop(0)
        module_logger.debug(
            "Imported app at app_path and cleaned up temporary modification to sys.path %s", self.app_path
        )

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Unload the imported module
        del sys.modules["app"]
        module_logger.debug("Deleted app from sys.modules and cleaned up (app_path %s)", self.app_path)

    @property
    def run(self):
        """Returns the unwrapped run function from app.py in the application's root directory"""
        return unwrap(self.app_module.run)



import concurrent.futures
import time
import google.api_core.exceptions

from octue import exceptions
from octue.resources.communication.google_pub_sub.service import OCTUE_NAMESPACE, Service
from octue.resources.communication.service_backends import GCPPubSubBackend
from octue.resources.manifest import Manifest
from tests import TEST_PROJECT_NAME
from tests.base import BaseTestCase


SERVER_WAIT_TIME = 5


class MockAnalysis:
    output_values = "Hello! It worked!"
    output_manifest = None


class DifferentMockAnalysis:
    output_values = "This is another successful analysis."
    output_manifest = None


class MockAnalysisWithOutputManifest:
    output_values = "This is an analysis with an empty output manifest."
    output_manifest = Manifest()


class TestService(BaseTestCase):
    """Some of these tests require a connection to either a real Google Pub/Sub instance on Google Cloud Platform
    (GCP), or a local emulator."""

    BACKEND = GCPPubSubBackend(
        project_name=TEST_PROJECT_NAME, credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS"
    )

    @staticmethod
    def ask_question_and_wait_for_answer(asking_service, responding_service, input_values, input_manifest):
        """ Get an asking service to ask a question to a responding service and wait for the answer. """
        subscription, _ = asking_service.ask(responding_service.id, input_values, input_manifest)
        return asking_service.wait_for_answer(subscription)

    @staticmethod
    def _delete_topics_and_subscriptions(*services):
        """Delete the topics and subscriptions created for each service. This method is necessary as it is difficult to
        end the threads running serving services gracefully, which would delete the topics and subscriptions if asked.
        """
        for service in services:
            topic_path = service.publisher.topic_path(service.backend.project_name, f"{OCTUE_NAMESPACE}.{service.id}")
            try:
                service.publisher.delete_topic(topic=topic_path)
            except google.api_core.exceptions.NotFound:
                pass

            subscription_path = service.subscriber.subscription_path(
                service.backend.project_name, f"{OCTUE_NAMESPACE}.{service.id}"
            )

            try:
                service.subscriber.delete_subscription(subscription=subscription_path)
            except (google.api_core.exceptions.NotFound, ValueError):
                pass

    def _shutdown_executor_and_clear_threads(self, executor):
        """Shut down a concurrent.futures executor by clearing its threads. This cuts loose infinitely-waiting threads,
        allowing the test to finish.
        """
        executor._threads.clear()
        executor.shutdown()
        concurrent.futures.thread._threads_queues.clear()

    def test_repr(self):
        """ Test that services are represented as a string correctly. """
        asking_service = Service(backend=self.BACKEND)
        self.assertEqual(repr(asking_service), f"<Service({asking_service.name!r})>")

    def test_serve_with_timeout(self):
        """ Test that a serving service only serves for as long as its timeout. """
        responding_service = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            start_time = time.perf_counter()
            responding_future = executor.submit(
                responding_service.serve,
                timeout=10,
                delete_topic_and_subscription_on_exit=True,
            )

            responding_future.result()
            self.assertTrue(time.perf_counter() - start_time < 20)

        self._delete_topics_and_subscriptions(responding_service)

    def test_ask_on_non_existent_service_results_in_error(self):
        """Test that trying to ask a question to a non-existent service (i.e. one without a topic in Google Pub/Sub)
        results in an error."""
        with self.assertRaises(exceptions.ServiceNotFound):
            Service(backend=self.BACKEND).ask(service_id="hello", input_values=[1, 2, 3, 4])

    def test_ask(self):
        """ Test that a service can ask a question to another service that is serving and receive an answer. """
        asking_service = Service(backend=self.BACKEND)
        responding_service = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            executor.submit(responding_service.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding service to be ready to answer.

            asker_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=asking_service,
                responding_service=responding_service,
                input_values={},
                input_manifest=None,
            )

            self.assertEqual(
                asker_future.result(),
                {"output_values": MockAnalysis.output_values, "output_manifest": MockAnalysis.output_manifest},
            )

            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(responding_service)

    def test_ask_with_input_manifest(self):
        """Test that a service can ask a question including an input_manifest to another service that is serving and
        receive an answer.
        """
        asking_service = Service(backend=self.BACKEND)
        responding_service = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            executor.submit(responding_service.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding service to be ready to answer.

            asker_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=asking_service,
                responding_service=responding_service,
                input_values={},
                input_manifest=Manifest(),
            )

            self.assertEqual(
                asker_future.result(),
                {"output_values": MockAnalysis.output_values, "output_manifest": MockAnalysis.output_manifest},
            )

            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(responding_service)

    def test_ask_with_output_manifest(self):
        """ Test that a service can receive an output manifest as part of the answer to a question. """
        asking_service = Service(backend=self.BACKEND)
        responding_service = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysisWithOutputManifest())

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            executor.submit(responding_service.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding service to be ready to answer.

            asker_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=asking_service,
                responding_service=responding_service,
                input_values={},
                input_manifest=None,
            )

            answer = asker_future.result()
            self.assertEqual(answer["output_values"], MockAnalysisWithOutputManifest.output_values)
            self.assertEqual(answer["output_manifest"].id, MockAnalysisWithOutputManifest.output_manifest.id)
            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(responding_service)

    def test_service_can_ask_multiple_questions(self):
        """ Test that a service can ask multiple questions to the same server and expect replies to them all. """
        asking_service = Service(backend=self.BACKEND)
        responding_service = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())

        with concurrent.futures.ThreadPoolExecutor(max_workers=7) as executor:
            executor.submit(responding_service.serve)
            futures = []

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding service to be ready to answer.

            for i in range(5):
                futures.append(
                    executor.submit(
                        self.ask_question_and_wait_for_answer,
                        asking_service=asking_service,
                        responding_service=responding_service,
                        input_values={},
                        input_manifest=None,
                    )
                )

            self._shutdown_executor_and_clear_threads(executor)

        for answer in concurrent.futures.as_completed(futures):
            self.assertEqual(
                answer.result(),
                {"output_values": MockAnalysis.output_values, "output_manifest": MockAnalysis.output_manifest},
            )

        self._delete_topics_and_subscriptions(responding_service)

    def test_service_can_ask_questions_to_multiple_servers(self):
        """ Test that a service can ask questions to different servers and expect replies to them all. """
        asking_service = Service(backend=self.BACKEND)

        responding_service_1 = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())
        responding_service_2 = self.make_new_server(self.BACKEND, run_function_returnee=DifferentMockAnalysis())

        with concurrent.futures.ThreadPoolExecutor(max_workers=7) as executor:
            executor.submit(responding_service_1.serve)
            executor.submit(responding_service_2.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding services to be ready to answer.

            first_question_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=asking_service,
                responding_service=responding_service_1,
                input_values={},
                input_manifest=None,
            )

            second_question_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=asking_service,
                responding_service=responding_service_2,
                input_values={},
                input_manifest=None,
            )

            self.assertEqual(
                first_question_future.result(),
                {"output_values": MockAnalysis.output_values, "output_manifest": MockAnalysis.output_manifest},
            )

            self.assertEqual(
                second_question_future.result(),
                {
                    "output_values": DifferentMockAnalysis.output_values,
                    "output_manifest": DifferentMockAnalysis.output_manifest,
                },
            )

            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(responding_service_1, responding_service_2)

    def test_server_can_ask_its_own_child_questions(self):
        """Test that a child can contact its own child while answering a question from a parent."""
        child_of_child = self.make_new_server(self.BACKEND, run_function_returnee=DifferentMockAnalysis())

        def child_run_function(input_values, input_manifest):
            service = Service(backend=self.BACKEND)
            subscription, _ = service.ask(service_id=child_of_child.id, input_values=input_values)

            mock_analysis = MockAnalysis()
            mock_analysis.output_values = {input_values["question"]: service.wait_for_answer(subscription)}
            return mock_analysis

        parent = Service(backend=self.BACKEND)
        child = Service(backend=self.BACKEND, run_function=child_run_function)

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            executor.submit(child.serve)
            executor.submit(child_of_child.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding services to be ready to answer.

            asker_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=parent,
                responding_service=child,
                input_values={"question": "What does the child of the child say?"},
                input_manifest=None,
            )

            self.assertEqual(
                asker_future.result(),
                {
                    "output_values": {
                        "What does the child of the child say?": {
                            "output_values": DifferentMockAnalysis.output_values,
                            "output_manifest": DifferentMockAnalysis.output_manifest,
                        }
                    },
                    "output_manifest": None,
                },
            )

            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(child, child_of_child)

    def test_server_can_ask_its_own_children_questions(self):
        """Test that a child can contact more than one of its own children while answering a question from a parent."""
        first_child_of_child = self.make_new_server(self.BACKEND, run_function_returnee=DifferentMockAnalysis())

        second_child_of_child = self.make_new_server(self.BACKEND, run_function_returnee=MockAnalysis())

        def child_run_function(input_values, input_manifest):
            child_service = Service(backend=self.BACKEND)
            subscription_1, _ = child_service.ask(service_id=first_child_of_child.id, input_values=input_values)
            subscription_2, _ = child_service.ask(service_id=second_child_of_child.id, input_values=input_values)

            mock_analysis = MockAnalysis()
            mock_analysis.output_values = {
                "first_child_of_child": child_service.wait_for_answer(subscription_1),
                "second_child_of_child": child_service.wait_for_answer(subscription_2),
            }

            return mock_analysis

        parent = Service(backend=self.BACKEND)
        child = Service(backend=self.BACKEND, run_function=child_run_function)

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            executor.submit(child.serve)
            executor.submit(first_child_of_child.serve)
            executor.submit(second_child_of_child.serve)

            time.sleep(SERVER_WAIT_TIME)  # Wait for the responding services to be ready to answer.

            asker_future = executor.submit(
                self.ask_question_and_wait_for_answer,
                asking_service=parent,
                responding_service=child,
                input_values={"question": "What does the child of the child say?"},
                input_manifest=None,
            )

            self.assertEqual(
                asker_future.result(),
                {
                    "output_values": {
                        "first_child_of_child": {
                            "output_values": DifferentMockAnalysis.output_values,
                            "output_manifest": DifferentMockAnalysis.output_manifest,
                        },
                        "second_child_of_child": {
                            "output_values": MockAnalysis.output_values,
                            "output_manifest": MockAnalysis.output_manifest,
                        },
                    },
                    "output_manifest": None,
                },
            )

            self._shutdown_executor_and_clear_threads(executor)

        self._delete_topics_and_subscriptions(child, first_child_of_child, second_child_of_child)



import json

from octue.utils.encoders import OctueJSONEncoder


class Serialisable:
    """Mixin class to make resources serialisable to JSON.

    Objects must have a `.logger` and a `.id` property

    """

    _SERIALISE_FIELDS = None
    _EXCLUDE_SERIALISE_FIELDS = ("logger",)

    def __init__(self, *args, **kwargs):
        """Constructor for serialisable mixin"""
        # Ensure it passes construction arguments up the chain
        super().__init__(*args, **kwargs)

    def to_file(self, file_name, **kwargs):
        """Write to a JSON file

        :parameter file_name:  file to write to, including relative or absolute path and .json extension
        :type file_name: path-like
        """
        self.logger.debug("Writing %s %s to file %s", self.__class__.__name__, self.id, file_name)
        with open(file_name, "w") as fp:
            fp.write(self.serialise(**kwargs, to_string=True))

    def serialise(self, to_string=False, **kwargs):
        """Serialise into a primitive dict or JSON string

        Serialises all non-private and non-protected attributes except for 'logger', unless the subclass has a
        `_serialise_fields` tuple of the attribute names to serialise. For example:
        ```
        class MyThing(Serialisable):
            _serialise_fields = ("a",)
            def __init__(self):
                self.a = 1
                self.b = 2

        MyThing().serialise()
        {"a": 1}
        ```

        By default, serialises using the OctueJSONEncoder, and will sort keys as well as format and indent
        automatically. Additional keyword arguments will be passed to ``json.dumps()`` to enable full override
        of formatting options

        :return: json string or dict contianing a serialised / primitive version of the resource.
        :rtype: str, dict
        """
        self.logger.debug("Serialising %s %s", self.__class__.__name__, self.id)

        # Get all non-private and non-protected attributes except those excluded specifically
        names_of_attributes_to_serialise = self._SERIALISE_FIELDS or (
            field_name
            for field_name in dir(self)
            if (
                field_name not in self._EXCLUDE_SERIALISE_FIELDS
                and (field_name[:1] != "_")
                and (type(getattr(self, field_name, "")).__name__ != "method")
            )
        )

        self_as_primitive = {}
        for name in names_of_attributes_to_serialise:
            attribute = getattr(self, name, None)
            # Serialise sets as sorted list (JSON doesn't support sets).
            self_as_primitive[name] = sorted(attribute) if isinstance(attribute, set) else attribute

        # TODO this conversion backward-and-forward is very inefficient but allows us to use the same encoder for
        #  converting the object to a dict as to strings, which ensures that nested attributes are also cast to
        #  primitive using their serialise() method. A more performant method would be to implement an encoder which
        #  returns python primitives, not strings. The reason we do this is to validate outbound information the same
        #  way as we validate incoming.
        string = json.dumps(self_as_primitive, cls=OctueJSONEncoder, sort_keys=True, indent=4, **kwargs)

        if to_string:
            return string

        return json.loads(string)



import collections.abc
import numbers

from octue import exceptions


IS_FILTER_ACTIONS = {
    "is": lambda item, filter_value: item is filter_value,
    "is_not": lambda item, filter_value: item is not filter_value,
}

EQUALS_FILTER_ACTIONS = {
    "equals": lambda item, filter_value: filter_value == item,
    "not_equals": lambda item, filter_value: filter_value != item,
}

COMPARISON_FILTER_ACTIONS = {
    "lt": lambda item, filter_value: item < filter_value,
    "lte": lambda item, filter_value: item <= filter_value,
    "gt": lambda item, filter_value: item > filter_value,
    "gte": lambda item, filter_value: item >= filter_value,
}

CONTAINS_FILTER_ACTIONS = {
    "contains": lambda item, filter_value: filter_value in item,
    "not_contains": lambda item, filter_value: filter_value not in item,
}

ICONTAINS_FILTER_ACTIONS = {
    "icontains": lambda item, filter_value: filter_value.lower() in item.lower(),
    "not_icontains": lambda item, filter_value: filter_value.lower() not in item.lower(),
}


# Filters for specific types e.g. list or int.
TYPE_FILTERS = {
    "bool": IS_FILTER_ACTIONS,
    "str": {
        "iequals": lambda item, filter_value: filter_value.lower() == item.lower(),
        "not_iequals": lambda item, filter_value: filter_value.lower() != item.lower(),
        "starts_with": lambda item, filter_value: item.startswith(filter_value),
        "not_starts_with": lambda item, filter_value: not item.startswith(filter_value),
        "ends_with": lambda item, filter_value: item.endswith(filter_value),
        "not_ends_with": lambda item, filter_value: not item.endswith(filter_value),
        **EQUALS_FILTER_ACTIONS,
        **COMPARISON_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **ICONTAINS_FILTER_ACTIONS,
    },
    "NoneType": IS_FILTER_ACTIONS,
    "TagSet": {
        "any_tag_contains": lambda item, filter_value: item.any_tag_contains(filter_value),
        "not_any_tag_contains": lambda item, filter_value: not item.any_tag_contains(filter_value),
        "any_tag_starts_with": lambda item, filter_value: item.any_tag_starts_with(filter_value),
        "not_any_tag_starts_with": lambda item, filter_value: not item.any_tag_starts_with(filter_value),
        "any_tag_ends_with": lambda item, filter_value: item.any_tag_ends_with(filter_value),
        "not_any_tag_ends_with": lambda item, filter_value: not item.any_tag_ends_with(filter_value),
        **EQUALS_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
    },
}

# Filters for interfaces e.g. iterables or numbers.
INTERFACE_FILTERS = {
    numbers.Number: {**EQUALS_FILTER_ACTIONS, **COMPARISON_FILTER_ACTIONS, **IS_FILTER_ACTIONS},
    collections.abc.Iterable: {
        **EQUALS_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **ICONTAINS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
    },
}


class Filterable:
    def satisfies(self, filter_name, filter_value):
        """ Check that the instance satisfies the given filter for the given filter value. """
        attribute_name, filter_action = self._split_filter_name(filter_name)

        try:
            attribute = getattr(self, attribute_name)
        except AttributeError:
            raise AttributeError(f"An attribute named {attribute_name!r} does not exist on {self!r}.")

        filter_ = self._get_filter(attribute, filter_action)
        return filter_(attribute, filter_value)

    def _split_filter_name(self, filter_name):
        """Split the filter name into the attribute name and filter action, raising an error if it the attribute name
        and filter action aren't delimited by a double underscore i.e. "__".
        """
        try:
            attribute_name, filter_action = filter_name.split("__", 1)
        except ValueError:
            raise exceptions.InvalidInputException(
                f"Invalid filter name {filter_name!r}. Filter names should be in the form "
                f"'<attribute_name>__<filter_kind>'."
            )

        return attribute_name, filter_action

    def _get_filter(self, attribute, filter_action):
        """Get the filter for the attribute and filter action, raising an error if there is no filter action of that
        name.
        """
        try:
            return self._get_filter_actions_for_attribute(attribute)[filter_action]

        except KeyError as error:
            attribute_type = type(attribute)
            raise exceptions.InvalidInputException(
                f"There is no filter called {error.args[0]!r} for attributes of type {attribute_type}. The options "
                f"are {self._get_filter_actions_for_attribute(attribute).keys()!r}"
            )

    def _get_filter_actions_for_attribute(self, attribute):
        """Get the possible filters for the given attribute based on its type or interface, raising an error if the
        attribute's type isn't supported (i.e. if there aren't any filters defined for it)."""
        try:
            return TYPE_FILTERS[type(attribute).__name__]

        except KeyError as error:
            # This allows handling of objects that conform to a certain interface (e.g. iterables) without needing the
            # specific type.
            for type_ in INTERFACE_FILTERS:
                if not isinstance(attribute, type_):
                    continue
                return INTERFACE_FILTERS[type_]

            raise exceptions.InvalidInputException(
                f"Attributes of type {error.args[0]} are not currently supported for filtering."
            )



import json
import tempfile
from unittest.mock import patch
import google.api_core.exceptions

from octue.utils.cloud import storage
from octue.utils.cloud.storage.client import GoogleCloudStorageClient
from tests import TEST_BUCKET_NAME, TEST_PROJECT_NAME
from tests.base import BaseTestCase


class TestUploadFileToGoogleCloud(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls.FILENAME = "my_file.txt"
        cls.storage_client = GoogleCloudStorageClient(project_name=TEST_PROJECT_NAME)

    def test_upload_and_download_file(self):
        """Test that a file can be uploaded to Google Cloud storage and downloaded again."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            upload_local_path = f"{temporary_directory}/{self.FILENAME}"

            with open(upload_local_path, "w") as f:
                f.write("This is a test upload.")

            self.storage_client.upload_file(
                local_path=upload_local_path,
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=self.FILENAME,
            )

            download_local_path = f"{temporary_directory}/{self.FILENAME}-download"

            self.storage_client.download_to_file(
                bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME, local_path=download_local_path
            )

            with open(download_local_path) as f:
                self.assertTrue("This is a test upload." in f.read())

    def test_upload_file_fails_if_checksum_is_not_correct(self):
        """Test that uploading a file fails if its checksum isn't the correct."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            upload_local_path = f"{temporary_directory}/{self.FILENAME}"

            with open(upload_local_path, "w") as f:
                f.write("This is a test upload.")

            with patch(
                "octue.utils.cloud.storage.client.GoogleCloudStorageClient._compute_crc32c_checksum",
                return_value="L3eGig==",
            ):

                with self.assertRaises(google.api_core.exceptions.BadRequest) as e:
                    self.storage_client.upload_file(
                        local_path=upload_local_path,
                        bucket_name=TEST_BUCKET_NAME,
                        path_in_bucket=self.FILENAME,
                    )

                self.assertTrue("doesn't match calculated CRC32C" in e.exception.message)

    def test_upload_from_string(self):
        """Test that a string can be uploaded to Google Cloud storage as a file and downloaded again."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            self.storage_client.upload_from_string(
                string=json.dumps({"height": 32}),
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=self.FILENAME,
            )

            download_local_path = f"{temporary_directory}/{self.FILENAME}-download"

            self.storage_client.download_to_file(
                bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME, local_path=download_local_path
            )

            with open(download_local_path) as f:
                self.assertTrue('{"height": 32}' in f.read())

    def test_upload_from_string_fails_if_checksum_is_not_correct(self):
        """Test that uploading a string fails if its checksum isn't the correct."""
        with patch(
            "octue.utils.cloud.storage.client.GoogleCloudStorageClient._compute_crc32c_checksum",
            return_value="L3eGig==",
        ):
            with self.assertRaises(google.api_core.exceptions.BadRequest) as e:
                self.storage_client.upload_from_string(
                    string=json.dumps({"height": 15}),
                    bucket_name=TEST_BUCKET_NAME,
                    path_in_bucket=self.FILENAME,
                )

            self.assertTrue("doesn't match calculated CRC32C" in e.exception.message)

    def test_download_as_string(self):
        """Test that a file can be uploaded to Google Cloud storage and downloaded as a string."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            upload_local_path = f"{temporary_directory}/{self.FILENAME}"

            with open(upload_local_path, "w") as f:
                f.write("This is a test upload.")

            self.storage_client.upload_file(
                local_path=upload_local_path,
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=self.FILENAME,
            )

        self.assertEqual(
            self.storage_client.download_as_string(bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME),
            "This is a test upload.",
        )

    def test_delete(self):
        """Test that a file can be deleted."""
        self.storage_client.upload_from_string(
            string=json.dumps({"height": 32}),
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket=self.FILENAME,
        )

        self.assertEqual(
            json.loads(
                self.storage_client.download_as_string(bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME),
            ),
            {"height": 32},
        )

        self.storage_client.delete(bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME)

        with self.assertRaises(google.api_core.exceptions.NotFound):
            self.storage_client.download_as_string(bucket_name=TEST_BUCKET_NAME, path_in_bucket=self.FILENAME)

    def test_scandir(self):
        """Test that Google Cloud storage "directories"' contents can be listed."""
        directory_path = storage.path.join("my", "path")

        self.storage_client.upload_from_string(
            string=json.dumps({"height": 32}),
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket=storage.path.join(directory_path, self.FILENAME),
        )

        contents = list(self.storage_client.scandir(bucket_name=TEST_BUCKET_NAME, directory_path=directory_path))
        self.assertEqual(len(contents), 1)
        self.assertEqual(contents[0].name, storage.path.join(directory_path, self.FILENAME))

    def test_scandir_with_empty_directory(self):
        """Test that an empty directory shows as such."""
        directory_path = storage.path.join("another", "path")
        contents = list(self.storage_client.scandir(bucket_name=TEST_BUCKET_NAME, directory_path=directory_path))
        self.assertEqual(len(contents), 0)

    def test_get_metadata(self):
        """Test that file metadata can be retrieved."""
        self.storage_client.upload_from_string(
            string=json.dumps({"height": 32}),
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket=self.FILENAME,
        )

        metadata = self.storage_client.get_metadata(
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket=self.FILENAME,
        )

        self.assertTrue(len(metadata) > 0)



import logging
import google.api_core.exceptions


logger = logging.getLogger(__name__)


class Topic:
    """A candidate topic to use with Google Pub/Sub. The topic represented by an instance of this class does not
    necessarily already exist on the Google Pub/Sub servers.
    """

    def __init__(self, name, namespace, service):
        if name.startswith(namespace):
            self.name = name
        else:
            self.name = f"{namespace}.{name}"

        self.service = service
        self.path = self.service.publisher.topic_path(service.backend.project_name, self.name)

    def __repr__(self):
        return f"<{type(self).__name__}({self.name})>"

    def create(self, allow_existing=False):
        """ Create a Google Pub/Sub topic that can be published to. """
        if not allow_existing:
            self.service.publisher.create_topic(name=self.path)
            self._log_creation()
            return

        try:
            self.service.publisher.create_topic(name=self.path)
        except google.api_core.exceptions.AlreadyExists:
            pass
        self._log_creation()

    def delete(self):
        """ Delete the topic from Google Pub/Sub. """
        self.service.publisher.delete_topic(topic=self.path)
        logger.debug("%r deleted topic %r.", self.service, self.path)

    def exists(self):
        """ Check if the topic exists on the Google Pub/Sub servers. """
        try:
            self.service.publisher.get_topic(topic=self.path)
        except google.api_core.exceptions.NotFound:
            return False
        return True

    def _log_creation(self):
        """ Log the creation of the topic. """
        logger.debug("%r created topic %r.", self.service, self.path)



import copy
import json
import os
import tempfile

from octue.resources import Datafile, Dataset, Manifest
from octue.utils.cloud import storage
from octue.utils.cloud.storage.client import GoogleCloudStorageClient
from tests import TEST_BUCKET_NAME
from tests.base import BaseTestCase


class TestManifest(BaseTestCase):
    def test_hash_value(self):
        """Test hashing a manifest with multiple datasets gives a hash of length 8."""
        manifest = self.create_valid_manifest()
        hash_ = manifest.hash_value
        self.assertTrue(isinstance(hash_, str))
        self.assertTrue(len(hash_) == 8)

    def test_hashes_for_the_same_manifest_are_the_same(self):
        """ Ensure the hashes for two manifests that are exactly the same are the same."""
        first_file = self.create_valid_manifest()
        second_file = copy.deepcopy(first_file)
        self.assertEqual(first_file.hash_value, second_file.hash_value)

    def test_to_cloud(self):
        """Test that a manifest can be uploaded to the cloud as a serialised JSON file of the Manifest instance. """
        project_name = "test-project"

        with tempfile.TemporaryDirectory() as temporary_directory:
            file_0_path = os.path.join(temporary_directory, "file_0.txt")
            file_1_path = os.path.join(temporary_directory, "file_1.txt")

            with open(file_0_path, "w") as f:
                f.write("[1, 2, 3]")

            with open(file_1_path, "w") as f:
                f.write("[4, 5, 6]")

            dataset = Dataset(
                name="my-dataset",
                files={
                    Datafile(timestamp=None, path=file_0_path, sequence=0, tags={"hello"}),
                    Datafile(timestamp=None, path=file_1_path, sequence=1, tags={"goodbye"}),
                },
            )

            manifest = Manifest(datasets=[dataset], keys={"my-dataset": 0})

            manifest.to_cloud(
                project_name,
                TEST_BUCKET_NAME,
                path_to_manifest_file=storage.path.join("blah", "manifest.json"),
            )

        persisted_manifest = json.loads(
            GoogleCloudStorageClient(project_name).download_as_string(
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=storage.path.join("blah", "manifest.json"),
            )
        )

        self.assertEqual(persisted_manifest["datasets"], ["gs://octue-test-bucket/blah/my-dataset"])
        self.assertEqual(persisted_manifest["keys"], {"my-dataset": 0})

    def test_to_cloud_without_storing_datasets(self):
        """Test that a manifest can be uploaded to the cloud as a serialised JSON file of the Manifest instance. """
        project_name = "test-project"

        with tempfile.TemporaryDirectory() as temporary_directory:
            file_0_path = os.path.join(temporary_directory, "file_0.txt")
            file_1_path = os.path.join(temporary_directory, "file_1.txt")

            with open(file_0_path, "w") as f:
                f.write("[1, 2, 3]")

            with open(file_1_path, "w") as f:
                f.write("[4, 5, 6]")

            dataset = Dataset(
                name="my-dataset",
                path=temporary_directory,
                files={
                    Datafile(timestamp=None, path=file_0_path, sequence=0, tags={"hello"}),
                    Datafile(timestamp=None, path=file_1_path, sequence=1, tags={"goodbye"}),
                },
            )

            manifest = Manifest(datasets=[dataset], keys={"my-dataset": 0})

            manifest.to_cloud(
                project_name,
                TEST_BUCKET_NAME,
                path_to_manifest_file=storage.path.join("my-manifests", "manifest.json"),
                store_datasets=False,
            )

        persisted_manifest = json.loads(
            GoogleCloudStorageClient(project_name).download_as_string(
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=storage.path.join("my-manifests", "manifest.json"),
            )
        )

        self.assertEqual(persisted_manifest["datasets"], [temporary_directory])
        self.assertEqual(persisted_manifest["keys"], {"my-dataset": 0})

    def test_from_cloud(self):
        """Test that a Manifest can be instantiated from the cloud."""
        project_name = "test-project"

        with tempfile.TemporaryDirectory() as temporary_directory:
            file_0_path = os.path.join(temporary_directory, "file_0.txt")
            file_1_path = os.path.join(temporary_directory, "file_1.txt")

            with open(file_0_path, "w") as f:
                f.write("[1, 2, 3]")

            with open(file_1_path, "w") as f:
                f.write("[4, 5, 6]")

            dataset = Dataset(
                name="my-dataset",
                files={
                    Datafile(timestamp=None, path=file_0_path, sequence=0, tags={"hello"}),
                    Datafile(timestamp=None, path=file_1_path, sequence=1, tags={"goodbye"}),
                },
            )

            manifest = Manifest(datasets=[dataset], keys={"my-dataset": 0})
            manifest.to_cloud(
                project_name,
                TEST_BUCKET_NAME,
                path_to_manifest_file=storage.path.join("my-directory", "manifest.json"),
            )

        persisted_manifest = Manifest.from_cloud(
            project_name=project_name,
            bucket_name=TEST_BUCKET_NAME,
            path_to_manifest_file=storage.path.join("my-directory", "manifest.json"),
        )

        self.assertEqual(persisted_manifest.path, f"gs://{TEST_BUCKET_NAME}/my-directory/manifest.json")
        self.assertEqual(persisted_manifest.id, manifest.id)
        self.assertEqual(persisted_manifest.hash_value, manifest.hash_value)
        self.assertEqual(persisted_manifest.keys, manifest.keys)
        self.assertEqual(
            {dataset.name for dataset in persisted_manifest.datasets}, {dataset.name for dataset in manifest.datasets}
        )

        for dataset in persisted_manifest.datasets:
            self.assertEqual(dataset.path, f"gs://{TEST_BUCKET_NAME}/my-directory/{dataset.name}")
            self.assertTrue(len(dataset.files), 2)
            self.assertTrue(all(isinstance(file, Datafile) for file in dataset.files))



import logging

from .cli import octue_cli
from .logging_handlers import LOG_FORMAT
from .runner import Runner


__all__ = "LOG_FORMAT", "octue_cli", "Runner"
package_logger = logging.getLogger(__name__)



import base64
import json
import uuid
from unittest import TestCase, mock

from octue.deployment.google import cloud_run
from octue.resources.communication.google_pub_sub.service import Service
from octue.resources.communication.service_backends import GCPPubSubBackend
from tests import TEST_PROJECT_NAME


cloud_run.app.testing = True


class TestCloudRun(TestCase):
    def test_post_to_index_with_no_payload_results_in_400_error(self):
        """Test that a 400 (bad request) error code is returned if no payload is sent to the Flask endpoint."""
        with cloud_run.app.test_client() as client:
            response = client.post("/")
            self.assertEqual(response.status_code, 400)

    def test_post_to_index_with_invalid_payload_results_in_400_error(self):
        """Test that a 400 (bad request) error code is returned if an invalid payload is sent to the Flask endpoint."""
        with cloud_run.app.test_client() as client:
            response = client.post("/", json={"some": "data"})
            self.assertEqual(response.status_code, 400)

            response = client.post("/", json={"message": "data"})
            self.assertEqual(response.status_code, 400)

    def test_post_to_index_with_valid_payload(self):
        """Test that the Flask endpoint returns a 204 (ok, no content) response to a valid payload."""
        with cloud_run.app.test_client() as client:
            with mock.patch("octue.deployment.google.cloud_run.answer_question"):

                response = client.post(
                    "/",
                    json={
                        "subscription": "projects/my-project/subscriptions/my-subscription",
                        "message": {
                            "data": base64.b64encode(
                                json.dumps({"input_values": [1, 2, 3], "input_manifest": None}).encode()
                            ).decode(),
                            "attributes": {"question_uuid": str(uuid.uuid4())},
                        },
                    },
                )

                self.assertEqual(response.status_code, 204)

    def test_cloud_run_integration(self):
        """Test that the Google Cloud Run integration works, providing a service that can be asked questions and send
        responses.
        """
        service_to_ask = "octue.services.009ea106-dc37-4521-a8cc-3e0836064334"
        asker = Service(backend=GCPPubSubBackend(project_name=TEST_PROJECT_NAME))
        subscription, _ = asker.ask(service_id=service_to_ask, input_values={"n_iterations": 3})
        answer = asker.wait_for_answer(subscription)
        self.assertEqual(answer, {"output_values": [1, 2, 3, 4, 5], "output_manifest": None})



import re
from functools import lru_cache

from octue.exceptions import InvalidTagException
from octue.mixins import Filterable
from octue.resources.filter_containers import FilterSet


TAG_PATTERN = re.compile(r"^$|^[a-z0-9][a-z0-9:\-]*(?<![:-])$")


class Tag(Filterable):
    """A tag starts and ends with a character in [a-z] or [0-9]. It can contain the colon discriminator or hyphens.
    Empty strings are also valid. More valid examples:
       system:32
       angry-marmaduke
       mega-man:torso:component:12
    """

    def __init__(self, name):
        self._name = self._clean(name)

    @property
    def name(self):
        return self._name

    @property
    @lru_cache(maxsize=1)
    def subtags(self):
        """ Return the subtags of the tag as a new TagSet (e.g. TagSet({'a', 'b', 'c'}) for the Tag('a:b:c'). """
        return TagSet({Tag(subtag_name) for subtag_name in (self.name.split(":"))})

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, Tag):
            return self.name == other.name
        return False

    def __lt__(self, other):
        if isinstance(other, str):
            return self.name < other
        elif isinstance(other, Tag):
            return self.name < other.name

    def __gt__(self, other):
        if isinstance(other, str):
            return self.name > other
        elif isinstance(other, Tag):
            return self.name > other.name

    def __hash__(self):
        """ Allow Tags to be contained in a set. """
        return hash(f"{type(self).__name__}{self.name}")

    def __contains__(self, item):
        return item in self.name

    def __repr__(self):
        return repr(self.name)

    def starts_with(self, value):
        """ Does the tag start with the given value? """
        return self.name.startswith(value)

    def ends_with(self, value):
        """ Does the tag end with the given value? """
        return self.name.endswith(value)

    @staticmethod
    def _clean(name):
        """ Ensure the tag name is a string and conforms to the tag regex pattern. """
        if not isinstance(name, str):
            raise InvalidTagException("Tags must be expressed as a string.")

        cleaned_name = name.strip()

        if not re.match(TAG_PATTERN, cleaned_name):
            raise InvalidTagException(
                f"Invalid tag '{cleaned_name}'. Tags must contain only characters 'a-z', '0-9', ':' and '-'. They must "
                f"not start with '-' or ':'."
            )

        return cleaned_name


class TagSet:
    """ Class to handle a set of tags as a string. """

    _FILTERSET_ATTRIBUTE = "tags"

    def __init__(self, tags=None, *args, **kwargs):
        """ Construct a TagSet. """
        # TODO Call the superclass with *args anad **kwargs, then update everything to using ResourceBase
        tags = tags or FilterSet()

        # Space delimited string of tag names.
        if isinstance(tags, str):
            self.tags = FilterSet(Tag(tag) for tag in tags.strip().split())

        elif isinstance(tags, TagSet):
            self.tags = FilterSet(tags.tags)

        # Tags can be some other iterable than a list, but each tag must be a Tag or string.
        elif hasattr(tags, "__iter__"):
            self.tags = FilterSet(tag if isinstance(tag, Tag) else Tag(tag) for tag in tags)

        else:
            raise InvalidTagException(
                "Tags must be expressed as a whitespace-delimited string or an iterable of strings or Tag instances."
            )

    def __str__(self):
        """ Serialise tags to a sorted list string. """
        return self.serialise()

    def __eq__(self, other):
        """ Does this TagSet have the same tags as another TagSet? """
        if not isinstance(other, TagSet):
            return False
        return self.tags == other.tags

    def __iter__(self):
        """ Iterate over the tags in the TagSet. """
        yield from self.tags

    def __len__(self):
        return len(self.tags)

    def __contains__(self, tag):
        """ Return True if any of the tags exactly matches value, allowing test like `if 'a' in TagSet('a b')`. """
        if isinstance(tag, str):
            return Tag(tag) in self.tags
        if isinstance(tag, Tag):
            return tag in self.tags

    def __repr__(self):
        return f"<TagSet({self.tags})>"

    def add_tags(self, *args):
        """Adds one or more new tag strings to the object tags. New tags will be cleaned and validated."""
        self.tags |= {Tag(arg) for arg in args}

    def get_subtags(self):
        """ Return a new TagSet instance with all the subtags. """
        return TagSet(subtag for tag in self for subtag in tag.subtags)

    def any_tag_starts_with(self, value):
        """ Implement a startswith method that returns true if any of the tags starts with value """
        return any(tag.starts_with(value) for tag in self)

    def any_tag_ends_with(self, value):
        """ Implement an endswith method that returns true if any of the tags endswith value. """
        return any(tag.ends_with(value) for tag in self)

    def any_tag_contains(self, value):
        """ Return True if any of the tags contains value. """
        return any(value in tag for tag in self)

    def serialise(self, to_string=True):
        """ Serialise tags to a sorted list string. """
        serialised_tags = sorted(tag.name for tag in self)

        if to_string:
            return str(serialised_tags)
        return serialised_tags



import base64
import json
import logging
from google.cloud import storage
from google.cloud.storage.constants import _DEFAULT_TIMEOUT
from google_crc32c import Checksum

from octue.utils.cloud.credentials import GCPCredentialsManager


logger = logging.getLogger(__name__)

OCTUE_MANAGED_CREDENTIALS = "octue-managed"


class GoogleCloudStorageClient:
    """A client for using Google Cloud Storage.

    :param str project_name:
    :param str|google.auth.credentials.Credentials credentials:
    :return None:
    """

    def __init__(self, project_name, credentials=OCTUE_MANAGED_CREDENTIALS):
        if credentials == OCTUE_MANAGED_CREDENTIALS:
            credentials = GCPCredentialsManager().get_credentials()
        else:
            credentials = credentials

        self.client = storage.Client(project=project_name, credentials=credentials)

    def upload_file(self, local_path, bucket_name, path_in_bucket, metadata=None, timeout=_DEFAULT_TIMEOUT):
        """Upload a local file to a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str local_path:
        :param str bucket_name:
        :param str path_in_bucket:
        :param dict metadata:
        :param float timeout:
        :return None:
        """
        blob = self._blob(bucket_name, path_in_bucket)

        with open(local_path) as f:
            blob.crc32c = self._compute_crc32c_checksum(f.read())

        blob.upload_from_filename(filename=local_path, timeout=timeout)
        self._update_metadata(blob, metadata)
        logger.info("Uploaded %r to Google Cloud at %r.", local_path, blob.public_url)

    def upload_from_string(self, string, bucket_name, path_in_bucket, metadata=None, timeout=_DEFAULT_TIMEOUT):
        """Upload serialised data in string form to a file in a Google Cloud bucket at
        gs://<bucket_name>/<path_in_bucket>.

        :param str string:
        :param str bucket_name:
        :param str path_in_bucket:
        :param dict metadata:
        :param float timeout:
        :return None:
        """
        blob = self._blob(bucket_name, path_in_bucket)
        blob.crc32c = self._compute_crc32c_checksum(string)

        blob.upload_from_string(data=string, timeout=timeout)
        self._update_metadata(blob, metadata)
        logger.info("Uploaded data to Google Cloud at %r.", blob.public_url)

    def download_to_file(self, bucket_name, path_in_bucket, local_path, timeout=_DEFAULT_TIMEOUT):
        """Download a file to a file from a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str bucket_name:
        :param str path_in_bucket:
        :param str local_path:
        :param float timeout:
        :return None:
        """
        blob = self._blob(bucket_name, path_in_bucket)
        blob.download_to_filename(local_path, timeout=timeout)
        logger.info("Downloaded %r from Google Cloud to %r.", blob.public_url, local_path)

    def download_as_string(self, bucket_name, path_in_bucket, timeout=_DEFAULT_TIMEOUT):
        """Download a file to a string from a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str bucket_name:
        :param str path_in_bucket:
        :param float timeout:
        :return str:
        """
        blob = self._blob(bucket_name, path_in_bucket)
        data = blob.download_as_bytes(timeout=timeout)
        logger.info("Downloaded %r from Google Cloud to as string.", blob.public_url)
        return data.decode()

    def get_metadata(self, bucket_name, path_in_bucket, timeout=_DEFAULT_TIMEOUT):
        """Get the metadata of the given file in the given bucket.

        :param str bucket_name:
        :param str path_in_bucket:
        :param float timeout:
        :return dict:
        """
        bucket = self.client.get_bucket(bucket_or_name=bucket_name)
        metadata = bucket.get_blob(blob_name=self._strip_leading_slash(path_in_bucket), timeout=timeout)._properties

        if metadata.get("metadata") is not None:
            metadata["metadata"] = {key: json.loads(value) for key, value in metadata["metadata"].items()}

        return metadata

    def delete(self, bucket_name, path_in_bucket, timeout=_DEFAULT_TIMEOUT):
        """Delete the given file from the given bucket.

        :param str bucket_name:
        :param str path_in_bucket:
        :param float timeout:
        :return None:
        """
        blob = self._blob(bucket_name, path_in_bucket)
        blob.delete(timeout=timeout)
        logger.info("Deleted %r from Google Cloud.", blob.public_url)

    def scandir(self, bucket_name, directory_path, filter=None, timeout=_DEFAULT_TIMEOUT):
        """Yield the blobs belonging to the given "directory" in the given bucket.

        :param str bucket_name:
        :param str directory_path:
        :param callable filter:
        :param float timeout:
        :yield google.cloud.storage.blob.Blob:
        """
        bucket = self.client.get_bucket(bucket_or_name=bucket_name)
        blobs = bucket.list_blobs(timeout=timeout)
        directory_path = self._strip_leading_slash(directory_path)

        if filter:
            return (blob for blob in blobs if blob.name.startswith(directory_path) and filter(blob))

        return (blob for blob in blobs if blob.name.startswith(directory_path))

    def _strip_leading_slash(self, path):
        """Strip the leading slash from a path.

        :param str path:
        :return str:
        """
        return path.lstrip("/")

    def _blob(self, bucket_name, path_in_bucket):
        """Instantiate a blob for the given bucket at the given path. Note that this is not synced up with Google Cloud.

        :param str bucket_name:
        :param str path_in_bucket:
        :return google.cloud.storage.blob.Blob:
        """
        bucket = self.client.get_bucket(bucket_or_name=bucket_name)
        return bucket.blob(blob_name=self._strip_leading_slash(path_in_bucket))

    def _compute_crc32c_checksum(self, string):
        """Compute the CRC32 checksum of the string.

        :param str string:
        :return str:
        """
        checksum = Checksum(string.encode())
        return base64.b64encode(checksum.digest()).decode("utf-8")

    def _update_metadata(self, blob, metadata):
        """Update the metadata for the given blob. Note that this is synced up with Google Cloud.

        :param google.cloud.storage.blob.Blob blob:
        :param dict metadata:
        :return None:
        """
        blob.metadata = self._encode_metadata(metadata or {})
        blob.patch()

    def _encode_metadata(self, metadata):
        """Encode metadata as a dictionary of JSON strings.

        :param dict metadata:
        :return dict:
        """
        if not isinstance(metadata, dict):
            raise TypeError(f"Metadata for Google Cloud storage should be a dictionary; received {metadata!r}")

        return {key: json.dumps(value) for key, value in metadata.items()}



from octue.resources.filter_containers import FilterSet
from octue.resources.tag import Tag, TagSet
from tests.base import BaseTestCase


class TestTag(BaseTestCase):
    def test_subtags(self):
        """ Test that subtags are correctly parsed from tags. """
        self.assertEqual(Tag("a:b:c").subtags, TagSet({Tag("a"), Tag("b"), Tag("c")}))

    def test_tag_comparison(self):
        """ Test that tags can be alphabetically compared. """
        self.assertTrue(Tag("a") < Tag("b"))
        self.assertTrue(Tag("b") > Tag("a"))
        self.assertTrue(Tag("a") != Tag("b"))
        self.assertTrue(Tag("a") == Tag("a"))

    def test_tag_comparison_with_strings(self):
        """ Test that tags can be alphabetically compared with strings in both directions. """
        self.assertTrue(Tag("a") < "b")
        self.assertTrue(Tag("b") > "a")
        self.assertTrue(Tag("a") != "b")
        self.assertTrue(Tag("a") == "a")
        self.assertTrue("b" > Tag("a"))
        self.assertTrue("a" < Tag("b"))
        self.assertTrue("b" != Tag("a"))
        self.assertTrue("a" == Tag("a"))

    def test_tags_compare_unequal_to_non_str_or_tag_types(self):
        """ Test that comparing for equality a Tag with a non-string-or-Tag type returns False. """
        self.assertFalse(Tag("a") == 1)
        self.assertTrue(Tag("a") != 1)

    def test_contains(self):
        """ Test that tags can be checked for containment. """
        self.assertIn("e", Tag("hello"))

    def test_starts_with(self):
        """ Test that the start of a tag can be checked. """
        self.assertTrue(Tag("hello").starts_with("h"))
        self.assertFalse(Tag("hello").starts_with("e"))

    def test_subtags_starts_with(self):
        """ Test that the start of subtags can be checked. """
        self.assertTrue(Tag("hello:world").subtags.any_tag_starts_with("w"))
        self.assertFalse(Tag("hello:world").subtags.any_tag_starts_with("e"))

    def test_ends_with(self):
        """ Test that the end of a tag can be checked. """
        self.assertTrue(Tag("hello").ends_with("o"))
        self.assertFalse(Tag("hello").ends_with("e"))

    def test_subtags_ends_with(self):
        """ Test that the end of subtags can be checked. """
        self.assertTrue(Tag("hello:world").subtags.any_tag_ends_with("o"))
        self.assertFalse(Tag("hello:world").subtags.any_tag_ends_with("e"))


class TestTagSet(BaseTestCase):
    TAG_SET = TagSet(tags="a b:c d:e:f")

    def test_instantiation_from_space_delimited_string(self):
        """ Test that a TagSet can be instantiated from a space-delimited string of tag names."""
        tag_set = TagSet(tags="a b:c d:e:f")
        self.assertEqual(tag_set.tags, FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))

    def test_instantiation_from_iterable_of_strings(self):
        """ Test that a TagSet can be instantiated from an iterable of strings."""
        tag_set = TagSet(tags=["a", "b:c", "d:e:f"])
        self.assertEqual(tag_set.tags, FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))

    def test_instantiation_from_iterable_of_tags(self):
        """ Test that a TagSet can be instantiated from an iterable of Tags."""
        tag_set = TagSet(tags=[Tag("a"), Tag("b:c"), Tag("d:e:f")])
        self.assertEqual(tag_set.tags, FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))

    def test_instantiation_from_filter_set_of_strings(self):
        """ Test that a TagSet can be instantiated from a FilterSet of strings."""
        tag_set = TagSet(tags=FilterSet({"a", "b:c", "d:e:f"}))
        self.assertEqual(tag_set.tags, FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))

    def test_instantiation_from_filter_set_of_tags(self):
        """ Test that a TagSet can be instantiated from a FilterSet of Tags."""
        tag_set = TagSet(tags=FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))
        self.assertEqual(tag_set.tags, FilterSet({Tag("a"), Tag("b:c"), Tag("d:e:f")}))

    def test_instantiation_from_tag_set(self):
        """ Test that a TagSet can be instantiated from another TagSet. """
        self.assertEqual(self.TAG_SET, TagSet(self.TAG_SET))

    def test_equality(self):
        """ Ensure two TagSets with the same tags compare equal. """
        self.assertTrue(self.TAG_SET == TagSet(tags="a b:c d:e:f"))

    def test_inequality(self):
        """ Ensure two TagSets with different tags compare unequal. """
        self.assertTrue(self.TAG_SET != TagSet(tags="a"))

    def test_non_tag_sets_compare_unequal_to_tag_sets(self):
        """ Ensure a TagSet and a non-TagSet compare unequal. """
        self.assertFalse(self.TAG_SET == "a")
        self.assertTrue(self.TAG_SET != "a")

    def test_iterating_over(self):
        """ Ensure a TagSet can be iterated over. """
        self.assertEqual(set(self.TAG_SET), {Tag("a"), Tag("b:c"), Tag("d:e:f")})

    def test_contains_with_string(self):
        """ Ensure we can check that a TagSet has a certain tag using a string form. """
        self.assertTrue("d:e:f" in self.TAG_SET)
        self.assertFalse("hello" in self.TAG_SET)

    def test_contains_with_tag(self):
        """ Ensure we can check that a TagSet has a certain tag. """
        self.assertTrue(Tag("d:e:f") in self.TAG_SET)
        self.assertFalse(Tag("hello") in self.TAG_SET)

    def test_contains_only_matches_full_tags(self):
        """ Test that the has_tag method only matches full tags (i.e. that it doesn't match subtags or parts of tags."""
        for tag in "a", "b:c", "d:e:f":
            self.assertTrue(tag in self.TAG_SET)

        for tag in "b", "c", "d", "e", "f":
            self.assertFalse(tag in self.TAG_SET)

    def test_get_subtags(self):
        """ Test subtags can be accessed as a new TagSet. """
        self.assertEqual(TagSet("meta:sys2:3456 blah").get_subtags(), TagSet("meta sys2 3456 blah"))

    def test_any_tag_starts_with(self):
        """ Ensure starts_with only checks the starts of tags, and doesn't check the starts of subtags. """
        for tag in "a", "b", "d":
            self.assertTrue(self.TAG_SET.any_tag_starts_with(tag))

        for tag in "c", "e", "f":
            self.assertFalse(self.TAG_SET.any_tag_starts_with(tag))

    def test_any_tag_ends_swith(self):
        """ Ensure ends_with doesn't check ends of subtags. """
        for tag in "a", "c", "f":
            self.assertTrue(self.TAG_SET.any_tag_ends_with(tag))

        for tag in "b", "d", "e":
            self.assertFalse(self.TAG_SET.any_tag_ends_with(tag))

    def test_any_tag_contains_searches_for_tags_and_subtags(self):
        """ Ensure tags and subtags can be searched for. """
        for tag in "a", "b", "d":
            self.assertTrue(self.TAG_SET.any_tag_contains(tag))

        for subtag in "c", "e", "f":
            self.assertTrue(self.TAG_SET.any_tag_contains(subtag))

    def test_filter(self):
        """ Test that tag sets can be filtered. """
        tag_set = TagSet(tags="tag1 tag2 meta:sys1:1234 meta:sys2:3456 meta:sys2:55")
        self.assertEqual(
            tag_set.tags.filter("name__starts_with", "meta"),
            FilterSet({Tag("meta:sys1:1234"), Tag("meta:sys2:3456"), Tag("meta:sys2:55")}),
        )

    def test_filter_chaining(self):
        """ Test that filters can be chained. """
        tag_set = TagSet(tags="tag1 tag2 meta:sys1:1234 meta:sys2:3456 meta:sys2:55")

        filtered_tags_1 = tag_set.tags.filter("name__starts_with", "meta")
        self.assertEqual(filtered_tags_1, TagSet("meta:sys1:1234 meta:sys2:3456 meta:sys2:55").tags)

        filtered_tags_2 = filtered_tags_1.filter("name__contains", "sys2")
        self.assertEqual(filtered_tags_2, TagSet("meta:sys2:3456 meta:sys2:55").tags)

        filtered_tags_3 = filtered_tags_1.filter("name__equals", "meta:sys2:55")
        self.assertEqual(filtered_tags_3, TagSet("meta:sys2:55").tags)

    def test_serialise(self):
        """ Ensure that TagSets are serialised to the string form of a list. """
        self.assertEqual(self.TAG_SET.serialise(), "['a', 'b:c', 'd:e:f']")

    def test_serialise_orders_tags(self):
        """ Ensure that TagSets are serialised to the string form of a list. """
        tag_set = TagSet("z hello a c:no")
        self.assertEqual(tag_set.serialise(), "['a', 'c:no', 'hello', 'z']")

    def test_str_is_equivalent_to_serialise(self):
        """ Test that calling `str` on a TagSet is equivalent to using the `serialise` method. """
        tag_set = TagSet("z hello a c:no")
        self.assertEqual(str(tag_set), tag_set.serialise())

    def test_repr(self):
        """ Test the representation of a TagSet appears as expected. """
        self.assertEqual(repr(self.TAG_SET), f"<TagSet({repr(self.TAG_SET.tags)})>")



from octue.exceptions import BackendNotFound
from octue.resources.communication.service_backends import get_backend
from tests.base import BaseTestCase


class TestServiceBackends(BaseTestCase):
    def test_error_raised_when_backend_type_does_not_exist(self):
        """ Test that an error is raised if the given backend name doesn't correspond to an existing backend type."""
        with self.assertRaises(BackendNotFound):
            get_backend("blahblah")

    def test_existing_backend_can_be_retrieved(self):
        """ Test that an existing backend can be retrieved. """
        backend = get_backend("GCPPubSubBackend")
        self.assertEqual(backend.__name__, "GCPPubSubBackend")



from octue.resources.tag import TagSet


class Taggable:
    """ A mixin class allowing objects to be tagged. """

    def __init__(self, *args, tags=None, **kwargs):
        """Constructor for Taggable mixins"""
        super().__init__(*args, **kwargs)
        self._tags = TagSet(tags)

    def add_tags(self, *args):
        """ Adds one or more new tag strings to the object tags. New tags will be cleaned and validated. """
        self._tags.add_tags(*args)

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        """ Overwrite any existing tag set and assign new tags. """
        self._tags = TagSet(tags)



from octue.utils.cloud import storage
from tests.base import BaseTestCase


class TestStorage(BaseTestCase):
    def test_join(self):
        """Test that paths can be joined correctly."""
        self.assertEqual(storage.path.join(), "")
        self.assertEqual(storage.path.join("a", "b", "c"), "a/b/c")
        self.assertEqual(storage.path.join("/a", "b", "c"), "/a/b/c")
        self.assertEqual(storage.path.join("", "nah", "blah"), "nah/blah")
        self.assertEqual(storage.path.join("", "nah", "", "blah"), "nah/blah")
        self.assertEqual(storage.path.join("a", "b", "/c"), "/c")
        self.assertEqual(storage.path.join("a", "b", "/c", "d"), "/c/d")

    def test_generate_gs_path(self):
        """Test that the `gs://` path can be generated correctly."""
        self.assertEqual(storage.path.generate_gs_path("my-bucket"), "gs://my-bucket")
        self.assertEqual(storage.path.generate_gs_path("my-bucket", "nah", "blah"), "gs://my-bucket/nah/blah")
        self.assertEqual(storage.path.generate_gs_path("my-bucket", "/nah", "blah"), "gs://my-bucket/nah/blah")

    def test_split_bucket_name_from_gs_path(self):
        """Test that the bucket name can be split from the path in a gs path."""
        self.assertEqual(
            storage.path.split_bucket_name_from_gs_path("gs://my-bucket/path/file.txt"), ("my-bucket", "path/file.txt")
        )
        self.assertEqual(storage.path.split_bucket_name_from_gs_path("gs://my-bucket"), ("my-bucket", ""))

    def test_strip_protocol_from_path(self):
        """Test that the `gs://` protocol can be stripped from a path."""
        self.assertEqual(storage.path.strip_protocol_from_path("gs://my-bucket"), "my-bucket")
        self.assertEqual(storage.path.strip_protocol_from_path("gs://my-bucket/"), "my-bucket/")
        self.assertEqual(storage.path.strip_protocol_from_path("gs://my-bucket/a/b/c"), "my-bucket/a/b/c")

    def test_relpath_with_gs_protocol(self):
        """Test that relative paths are calculated correctly for cloud paths."""
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="gs://my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="gs://my-bucket/"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="gs://my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="gs://my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="gs://my-bucket/a/b/c/"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="gs://my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="gs://my-bucket/a/d"), "../b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/d", start="gs://my-bucket/a/b/c"), "../../d")

    def test_relpath_with_gs_protocol_for_path(self):
        """Test that relative paths are calculated correctly for cloud paths."""
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="gs://my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="gs://my-bucket/"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="gs://my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="gs://my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="gs://my-bucket/a/b/c/"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="gs://my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="gs://my-bucket/a/d"), "../b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/d", start="gs://my-bucket/a/b/c"), "../../d")

    def test_relpath_without_gs_protocol_for_start(self):
        """Test that relative paths are calculated correctly for cloud paths."""
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="my-bucket/"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c", start="my-bucket/a/b/c/"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/b/c/", start="my-bucket/a/d"), "../b/c")
        self.assertEqual(storage.path.relpath(path="gs://my-bucket/a/d", start="my-bucket/a/b/c"), "../../d")

    def test_relpath_without_gs_protocol(self):
        """Test that relative paths are calculated correctly for cloud paths."""
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="my-bucket/"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="my-bucket"), "a/b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c", start="my-bucket/a/b/c/"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="my-bucket/a/b/c"), ".")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/b/c/", start="my-bucket/a/d"), "../b/c")
        self.assertEqual(storage.path.relpath(path="my-bucket/a/d", start="my-bucket/a/b/c"), "../../d")

    def test_split(self):
        """Test that cloud paths can be split properly."""
        self.assertEqual(storage.path.split("my-bucket/a/b/c.txt"), ("my-bucket/a/b", "c.txt"))

    def test_dirname(self):
        """Test that the name of the directory of the given path can be found."""
        self.assertEqual(storage.path.dirname("a/b/c"), "a/b")
        self.assertEqual(storage.path.dirname("a/b/c", name_only=True), "b")
        self.assertEqual(storage.path.dirname("/a/b/c"), "/a/b")
        self.assertEqual(storage.path.dirname("/a/b/c", name_only=True), "b")
        self.assertEqual(storage.path.dirname("a"), "")
        self.assertEqual(storage.path.dirname("/a"), "/")



import google.oauth2.service_account

from octue.utils.cloud.credentials import GCPCredentialsManager
from tests.base import BaseTestCase


class TestGCPCredentialsManager(BaseTestCase):
    def test_error_is_raise_if_no_environment_variable(self):
        """ Ensure an error is raised if the given environment variable can't be found. """
        with self.assertRaises(EnvironmentError):
            GCPCredentialsManager(environment_variable_name="heeby_jeeby-11111-33103")

    def test_credentials_can_be_loaded_from_file(self):
        """ Test that credentials can be loaded from the file at the path specified by the environment variable. """
        credentials = GCPCredentialsManager().get_credentials()
        self.assertEqual(type(credentials), google.oauth2.service_account.Credentials)



import logging
import google.api_core.exceptions


logger = logging.getLogger(__name__)


class Subscription:
    """A candidate subscription to use with Google Pub/Sub. The subscription represented by an instance of this class
    does not necessarily already exist on the Google Pub/Sub servers.
    """

    def __init__(self, name, topic, namespace, service):
        if name.startswith(namespace):
            self.name = name
        else:
            self.name = f"{namespace}.{name}"

        self.topic = topic
        self.service = service
        self.path = self.service.subscriber.subscription_path(self.service.backend.project_name, self.name)

    def __repr__(self):
        return f"<{type(self).__name__}({self.name})>"

    def create(self, allow_existing=False):
        """ Create a Google Pub/Sub subscription that can be subscribed to. """
        if not allow_existing:
            self.service.subscriber.create_subscription(topic=self.topic.path, name=self.path)
            self._log_creation()
            return

        try:
            self.service.subscriber.create_subscription(topic=self.topic.path, name=self.path)
        except google.api_core.exceptions.AlreadyExists:
            pass
        self._log_creation()

    def delete(self):
        """ Delete the subscription from Google Pub/Sub. """
        self.service.subscriber.delete_subscription(subscription=self.path)
        logger.debug("%r deleted subscription %r.", self.service, self.path)

    def _log_creation(self):
        """ Log the creation of the subscription. """
        logger.debug("%r created subscription %r.", self.service, self.path)



from octue import exceptions
from octue.mixins import Filterable
from octue.resources.filter_containers import FilterList, FilterSet
from tests.base import BaseTestCase


class Cat(Filterable):
    def __init__(self, name=None, previous_names=None, age=None):
        self.name = name
        self.previous_names = previous_names
        self.age = age


class TestFilterSet(BaseTestCase):
    def test_ordering_by_a_non_existent_attribute(self):
        """ Ensure an error is raised if ordering is attempted by a non-existent attribute. """
        filter_set = FilterSet([Cat(age=5), Cat(age=4), Cat(age=3)])
        with self.assertRaises(exceptions.InvalidInputException):
            filter_set.order_by("dog-likeness")

    def test_order_by_with_string_attribute(self):
        """ Test ordering a FilterSet by a string attribute returns an appropriately ordered FilterList. """
        cats = [Cat(name="Zorg"), Cat(name="James"), Cat(name="Princess Carolyn")]
        sorted_filter_set = FilterSet(cats).order_by("name")
        self.assertEqual(sorted_filter_set, FilterList([cats[1], cats[2], cats[0]]))

    def test_order_by_with_int_attribute(self):
        """ Test ordering a FilterSet by an integer attribute returns an appropriately ordered FilterList. """
        cats = [Cat(age=5), Cat(age=4), Cat(age=3)]
        sorted_filter_set = FilterSet(cats).order_by("age")
        self.assertEqual(sorted_filter_set, FilterList(reversed(cats)))

    def test_order_by_list_attribute(self):
        """ Test that ordering by list attributes orders by the size of the list. """
        cats = [Cat(previous_names=["Scatta", "Catta"]), Cat(previous_names=["Kitty"]), Cat(previous_names=[])]
        sorted_filter_set = FilterSet(cats).order_by("previous_names")
        self.assertEqual(sorted_filter_set, FilterList(reversed(cats)))

    def test_order_by_in_reverse(self):
        """ Test ordering in reverse works correctly. """
        cats = [Cat(age=5), Cat(age=3), Cat(age=4)]
        sorted_filter_set = FilterSet(cats).order_by("age", reverse=True)
        self.assertEqual(sorted_filter_set, FilterList([cats[0], cats[2], cats[1]]))



""" The names of backend classes must end with "Backend" to be registered as valid backends for the `Child` class.
Beyond that, a backend class can store whatever data it needs to authenticate and specify resources for the specific
backend it represents. Credentials should not be stored directly in a backend instance, but storing the names of
environment variables that store credentials is fine. For a backend to be valid, it must have a corresponding entry in
the "oneOf" field of the "backend" key of the children schema in `Twined`, which is located at
`twined/schema/children_schema.json`.
"""
from octue import exceptions


def get_backend(backend_name):
    if backend_name not in AVAILABLE_BACKENDS:
        raise exceptions.BackendNotFound(
            f"Backend with name {backend_name} not found. Available backends are {list(AVAILABLE_BACKENDS.keys())}"
        )

    return AVAILABLE_BACKENDS[backend_name]


class GCPPubSubBackend:
    """ A dataclass containing the details needed to use Google Cloud Platform Pub/Sub as a Service backend. """

    def __init__(self, project_name, credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS"):
        self.project_name = project_name
        self.credentials_environment_variable = credentials_environment_variable


AVAILABLE_BACKENDS = {key: value for key, value in locals().items() if key.endswith("Backend")}



import json
import logging

from octue.exceptions import InvalidInputException, InvalidManifestException
from octue.mixins import Hashable, Identifiable, Loggable, Pathable, Serialisable
from octue.utils.cloud import storage
from octue.utils.cloud.storage.client import GoogleCloudStorageClient
from .dataset import Dataset


module_logger = logging.getLogger(__name__)


class Manifest(Pathable, Serialisable, Loggable, Identifiable, Hashable):
    """A representation of a manifest, which can contain multiple datasets This is used to manage all files coming into
    (or leaving), a data service for an analysis at the configuration, input or output stage."""

    _ATTRIBUTES_TO_HASH = "datasets", "keys"

    def __init__(self, id=None, logger=None, path=None, path_from=None, datasets=None, keys=None, **kwargs):
        """Construct a Manifest"""
        super().__init__(id=id, logger=logger, path=path, path_from=path_from)

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then tidy up this hackjob. Also need to allow Pathables to update ownership
        #  (because decoders work from the bottom of the tree upwards, not top-down)

        datasets = datasets or []
        self.keys = keys or {}

        # TODO we need to add keys to the manifest file schema in twined so that we know what dataset(s) map to what keys
        #  In the meantime, we enforce at this level that keys will match
        if len(self.keys) != len(datasets):
            raise InvalidManifestException(
                f"Manifest instantiated with {len(self.keys)} keys, and {len(datasets)} datasets... keys must match datasets!"
            )

        # Sort the keys by the dataset index so we have a list of keys in the same order as the dataset list.
        # We'll use this to name the dataset folders
        key_list = [key for key, value in sorted(self.keys.items(), key=lambda item: item[1])]

        # Instantiate the datasets if not already done
        self.datasets = []
        for key, dataset in zip(key_list, datasets):
            if isinstance(dataset, Dataset):
                self.datasets.append(dataset)
            else:
                self.datasets.append(Dataset(**dataset, path=key, path_from=self))

        # Instantiate the rest of everything!
        vars(self).update(**kwargs)

    @classmethod
    def from_cloud(cls, project_name, bucket_name, path_to_manifest_file):
        """Instantiate a Manifest from Google Cloud storage.

        :param str project_name:
        :param str bucket_name:
        :param str path_to_manifest_file:
        :return Dataset:
        """
        storage_client = GoogleCloudStorageClient(project_name=project_name)

        serialised_manifest = json.loads(
            storage_client.download_as_string(bucket_name=bucket_name, path_in_bucket=path_to_manifest_file)
        )

        datasets = []

        for dataset in serialised_manifest["datasets"]:
            dataset_bucket_name, path = storage.path.split_bucket_name_from_gs_path(dataset)

            datasets.append(
                Dataset.from_cloud(
                    project_name=project_name, bucket_name=dataset_bucket_name, path_to_dataset_directory=path
                )
            )

        return Manifest(
            id=serialised_manifest["id"],
            path=storage.path.generate_gs_path(bucket_name, path_to_manifest_file),
            hash_value=serialised_manifest["hash_value"],
            datasets=datasets,
            keys=serialised_manifest["keys"],
        )

    def to_cloud(self, project_name, bucket_name, path_to_manifest_file, store_datasets=True):
        """Upload a manifest to a cloud location, optionally uploading its datasets into the same directory.

        :param str project_name:
        :param str bucket_name:
        :param str path_to_manifest_file:
        :param bool store_datasets: if True, upload datasets to same directory as manifest file
        :return str: gs:// path for manifest file
        """
        datasets = []
        output_directory = storage.path.dirname(path_to_manifest_file)

        for dataset in self.datasets:

            if store_datasets:
                dataset_path = dataset.to_cloud(project_name, bucket_name, output_directory=output_directory)
                datasets.append(dataset_path)
            else:
                datasets.append(dataset.absolute_path)

        serialised_manifest = self.serialise()
        serialised_manifest["datasets"] = sorted(datasets)
        del serialised_manifest["absolute_path"]
        del serialised_manifest["path"]

        GoogleCloudStorageClient(project_name=project_name).upload_from_string(
            string=json.dumps(serialised_manifest),
            bucket_name=bucket_name,
            path_in_bucket=path_to_manifest_file,
        )

        return storage.path.generate_gs_path(bucket_name, path_to_manifest_file)

    def get_dataset(self, key):
        """Gets a dataset by its key name (as defined in the twine)

        :return Dataset: Dataset selected by its key
        """
        idx = self.keys.get(key, None)
        if idx is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset '{key}' from Manifest. Allowable keys are: {list(self.keys.keys())}"
            )

        return self.datasets[idx]

    def prepare(self, data):
        """Prepare new manifest from a manifest_spec"""
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for idx, dataset_spec in enumerate(data):

            self.keys[dataset_spec["key"]] = idx
            # TODO generate a unique name based on the filter key, tag datasets so that the tag filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets.append(Dataset(logger=self.logger, path_from=self, path=dataset_spec["key"]))

        return self

    @classmethod
    def deserialise(cls, serialised_manifest, from_string=False):
        """ Deserialise a Manifest from a dictionary. """
        if from_string:
            serialised_manifest = json.loads(serialised_manifest)

        return cls(
            id=serialised_manifest["id"],
            datasets=serialised_manifest["datasets"],
            keys=serialised_manifest["keys"],
            path=serialised_manifest["path"],
        )



import json
import logging
import os
from google.oauth2 import service_account


logger = logging.getLogger(__name__)


class GCPCredentialsManager:
    """A credentials manager for Google Cloud Platform (GCP) that takes a JSON string from an environment variable and
    writes it to a file in a temporary directory.
    """

    def __init__(self, environment_variable_name="GOOGLE_APPLICATION_CREDENTIALS"):
        self.environment_variable_name = environment_variable_name

        try:
            self.environment_variable_value = os.environ[self.environment_variable_name]
        except KeyError:
            raise EnvironmentError(f"There is no environment variable called {self.environment_variable_name}.")

    def get_credentials(self):
        """Get the Google OAUTH2 service account credentials for which the environment variable value is either the
        filename containing them or a JSON string version of them."""

        # Check that the environment variable refers to a valid *and* real path.
        if os.path.exists(self.environment_variable_value):
            return self._get_credentials_from_file()

        # If it doesn't, assume that it's the credentials file as a JSON string.
        return self._get_credentials_from_string()

    def _get_credentials_from_file(self):
        with open(self.environment_variable_value) as f:
            credentials = json.load(f)

        logger.debug("GCP credentials read from file.")
        return service_account.Credentials.from_service_account_info(credentials)

    def _get_credentials_from_string(self):
        credentials = json.loads(self.environment_variable_value)
        logger.debug("GCP credentials loaded from string.")
        return service_account.Credentials.from_service_account_info(credentials)



import copy
import json
import os
import tempfile
import uuid

from octue import exceptions
from octue.mixins import MixinBase, Pathable
from octue.resources import Datafile
from octue.resources.tag import TagSet
from octue.utils.cloud.storage.client import GoogleCloudStorageClient
from tests import TEST_BUCKET_NAME, TEST_PROJECT_NAME
from ..base import BaseTestCase


class MyPathable(Pathable, MixinBase):
    pass


class DatafileTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.path_from = MyPathable(path=os.path.join(self.data_path, "basic_files", "configuration", "test-dataset"))
        self.path = os.path.join("path-within-dataset", "a_test_file.csv")

    def create_valid_datafile(self):
        return Datafile(timestamp=None, path_from=self.path_from, path=self.path, skip_checks=False)

    def test_instantiates(self):
        """Ensures a Datafile instantiates using only a path and generates a uuid ID"""
        df = Datafile(timestamp=None, path="a_path")
        self.assertTrue(isinstance(df.id, str))
        self.assertEqual(type(uuid.UUID(df.id)), uuid.UUID)
        self.assertIsNone(df.sequence)
        self.assertEqual(0, df.cluster)

    def test_path_argument_required(self):
        """Ensures instantiation without a path will fail"""
        with self.assertRaises(exceptions.InvalidInputException) as error:
            Datafile(timestamp=None)

        self.assertIn("You must supply a valid 'path' for a Datafile", error.exception.args[0])

    def test_gt(self):
        """Test that datafiles can be ordered using the greater-than operator."""
        a = Datafile(timestamp=None, path="a_path")
        b = Datafile(timestamp=None, path="b_path")
        self.assertTrue(a < b)

    def test_gt_with_wrong_type(self):
        """Test that datafiles cannot be ordered compared to other types."""
        with self.assertRaises(TypeError):
            Datafile(timestamp=None, path="a_path") < "hello"

    def test_lt(self):
        """Test that datafiles can be ordered using the less-than operator."""
        a = Datafile(timestamp=None, path="a_path")
        b = Datafile(timestamp=None, path="b_path")
        self.assertTrue(b > a)

    def test_lt_with_wrong_type(self):
        """Test that datafiles cannot be ordered compared to other types."""
        with self.assertRaises(TypeError):
            Datafile(timestamp=None, path="a_path") > "hello"

    def test_checks_fail_when_file_doesnt_exist(self):
        path = "not_a_real_file.csv"
        with self.assertRaises(exceptions.FileNotFoundException) as error:
            Datafile(timestamp=None, path=path, skip_checks=False)
        self.assertIn("No file found at", error.exception.args[0])

    def test_conflicting_extension_fails_check(self):
        with self.assertRaises(exceptions.InvalidInputException) as error:
            Datafile(timestamp=None, path_from=self.path_from, path=self.path, skip_checks=False, extension="notcsv")

        self.assertIn("Extension provided (notcsv) does not match file extension", error.exception.args[0])

    def test_file_attributes_accessible(self):
        """Ensures that its possible to set the sequence, cluster and timestamp"""
        df = self.create_valid_datafile()
        self.assertIsInstance(df.size_bytes, int)
        self.assertGreaterEqual(df._last_modified, 1598200190.5771205)
        self.assertEqual("a_test_file.csv", df.name)

        df.sequence = 2
        df.cluster = 0
        df.timestamp = 0

    def test_cannot_set_calculated_file_attributes(self):
        """Ensures that calculated attributes cannot be set"""
        df = self.create_valid_datafile()

        with self.assertRaises(AttributeError):
            df.size_bytes = 1

        with self.assertRaises(AttributeError):
            df._last_modified = 1000000000.5771205

    def test_repr(self):
        """ Test that Datafiles are represented as expected. """
        self.assertEqual(repr(self.create_valid_datafile()), "<Datafile('a_test_file.csv')>")

    def test_serialisable(self):
        """Ensures a datafile can serialise to json format"""
        df = self.create_valid_datafile()
        df_dict = df.serialise()

        for k in df_dict.keys():
            self.assertFalse(k.startswith("_"))

        for k in (
            "cluster",
            "extension",
            "id",
            "name",
            "path",
            "timestamp",
            "sequence",
            "size_bytes",
            "tags",
            "hash_value",
        ):
            self.assertIn(k, df_dict.keys())

    def test_hash_value(self):
        """Test hashing a datafile gives a hash of length 8."""
        hash_ = self.create_valid_datafile().hash_value
        self.assertTrue(isinstance(hash_, str))
        self.assertTrue(len(hash_) == 8)

    def test_hashes_for_the_same_datafile_are_the_same(self):
        """ Ensure the hashes for two datafiles that are exactly the same are the same."""
        first_file = self.create_valid_datafile()
        second_file = copy.deepcopy(first_file)
        self.assertEqual(first_file.hash_value, second_file.hash_value)

    def test_from_cloud_with_bare_file(self):
        """Test that a Datafile can be constructed from a file on Google Cloud storage with no custom metadata."""
        path_in_bucket = "file_to_upload.txt"

        GoogleCloudStorageClient(project_name=TEST_PROJECT_NAME).upload_from_string(
            string=json.dumps({"height": 32}),
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket=path_in_bucket,
        )

        datafile = Datafile.from_cloud(
            project_name=TEST_PROJECT_NAME, bucket_name=TEST_BUCKET_NAME, datafile_path=path_in_bucket, timestamp=None
        )

        self.assertEqual(datafile.path, f"gs://{TEST_BUCKET_NAME}/{path_in_bucket}")
        self.assertEqual(datafile.cluster, 0)
        self.assertEqual(datafile.sequence, None)
        self.assertEqual(datafile.tags, TagSet())
        self.assertTrue(isinstance(datafile.size_bytes, int))
        self.assertTrue(isinstance(datafile._last_modified, float))
        self.assertTrue(isinstance(datafile.hash_value, str))

    def test_from_cloud_with_datafile(self):
        """Test that a Datafile can be constructed from a file on Google Cloud storage with custom metadata."""
        path_in_bucket = "file_to_upload.txt"

        with tempfile.TemporaryDirectory() as temporary_directory:
            file_0_path = os.path.join(temporary_directory, "file_0.txt")

            with open(file_0_path, "w") as f:
                f.write("[1, 2, 3]")

            datafile = Datafile(
                timestamp=None, path=file_0_path, cluster=0, sequence=1, tags={"blah:shah:nah", "blib", "glib"}
            )
            datafile.to_cloud(
                project_name=TEST_PROJECT_NAME, bucket_name=TEST_BUCKET_NAME, path_in_bucket=path_in_bucket
            )

            persisted_datafile = Datafile.from_cloud(
                project_name=TEST_PROJECT_NAME, bucket_name=TEST_BUCKET_NAME, datafile_path=path_in_bucket
            )

            self.assertEqual(persisted_datafile.path, f"gs://{TEST_BUCKET_NAME}/{path_in_bucket}")
            self.assertEqual(persisted_datafile.id, datafile.id)
            self.assertEqual(persisted_datafile.hash_value, datafile.hash_value)
            self.assertEqual(persisted_datafile.cluster, datafile.cluster)
            self.assertEqual(persisted_datafile.sequence, datafile.sequence)
            self.assertEqual(persisted_datafile.tags, datafile.tags)
            self.assertEqual(persisted_datafile.size_bytes, datafile.size_bytes)
            self.assertTrue(isinstance(persisted_datafile._last_modified, float))



import os
import tempfile
import time
import unittest

from octue.utils.persistence import calculate_disk_usage, get_oldest_file_in_directory


def _create_file_of_size(path, size):
    """Create a file at the given path of the given size in bytes.

    :param str path:
    :param int size:
    :return None:
    """
    with open(path, "wb") as f:
        f.truncate(size)


class TestCalculateDiskUsage(unittest.TestCase):
    def test_calculate_disk_usage_with_single_file(self):
        """Test that the disk usage of a single file is calculated correctly."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            dummy_file_path = os.path.join(temporary_directory, "dummy_file")
            _create_file_of_size(dummy_file_path, 1)
            self.assertEqual(calculate_disk_usage(dummy_file_path), 1)

    def test_with_filter_satisfied(self):
        """Test that files meeting a filter are included in the calculated size."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            dummy_file_path = os.path.join(temporary_directory, "dummy_file")
            _create_file_of_size(dummy_file_path, 1)

            self.assertEqual(
                calculate_disk_usage(dummy_file_path, filter=lambda path: os.path.split(path)[-1].startswith("dummy")),
                1,
            )

    def test_with_filter_unsatisfied(self):
        """Test that files not meeting a filter are not included in the calculated size."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            dummy_file_path = os.path.join(temporary_directory, "dummy_file")
            _create_file_of_size(dummy_file_path, 1)

            self.assertEqual(
                calculate_disk_usage(dummy_file_path, filter=lambda path: os.path.split(path)[-1].startswith("hello")),
                0,
            )

    def test_calculate_disk_usage_with_shallow_directory(self):
        """Test that the disk usage of a directory with only files in is calculated correctly."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            for i in range(5):
                path = os.path.join(temporary_directory, f"dummy_file_{i}")
                _create_file_of_size(path, 1)

            self.assertEqual(calculate_disk_usage(temporary_directory), 5)

    def test_calculate_disk_usage_with_directory_of_directories(self):
        """Test that the disk usage of a directory of directories with files in is calculated correctly."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            for i in range(5):
                directory_path = os.path.join(temporary_directory, f"directory_{i}")
                os.mkdir(directory_path)

                for j in range(5):
                    file_path = os.path.join(directory_path, f"dummy_file_{j}")
                    _create_file_of_size(file_path, 1)

            self.assertEqual(calculate_disk_usage(temporary_directory), 25)

    def test_calculate_disk_usage_with_directory_of_directories_and_files(self):
        """Test that the disk usage of a directory of directories with files in and files next to the directories is
        calculated correctly.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:

            for i in range(5):
                directory_path = os.path.join(temporary_directory, f"directory_{i}")
                os.mkdir(directory_path)

                adjacent_file_path = os.path.join(temporary_directory, f"adjacent_dummy_file_{i}")
                _create_file_of_size(adjacent_file_path, 2)

                for j in range(5):
                    file_path = os.path.join(directory_path, f"dummy_file_{j}")
                    _create_file_of_size(file_path, 1)

            self.assertEqual(calculate_disk_usage(temporary_directory), 35)


class TestGetOldestFileInDirectory(unittest.TestCase):
    def test_oldest_file_is_retrieved(self):
        """Test that the path of the oldest file in a directory is retrieved."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            for i in range(5):
                _create_file_of_size(path=os.path.join(temporary_directory, f"file_{i}"), size=1)
                time.sleep(1)

            self.assertEqual(
                get_oldest_file_in_directory(temporary_directory), os.path.join(temporary_directory, "file_0")
            )

    def test_empty_directory_results_in_none(self):
        """Test that None is returned for an empty directory."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            self.assertEqual(get_oldest_file_in_directory(temporary_directory), None)

    def test_directories_ignored(self):
        """Test that directories are ignored."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            for i in range(5):
                os.mkdir(os.path.join(temporary_directory, f"directory_{i}"))

            self.assertEqual(get_oldest_file_in_directory(temporary_directory), None)

    def test_only_files_satisfying_filter_are_considered(self):
        """Test that only files satisfying the filter are considered when getting the oldest file."""
        with tempfile.TemporaryDirectory() as temporary_directory:

            _create_file_of_size(os.path.join(temporary_directory, "file_not_to_consider"), 1)
            _create_file_of_size(os.path.join(temporary_directory, "dummy_file"), 1)

            self.assertEqual(
                get_oldest_file_in_directory(
                    temporary_directory, filter=lambda path: os.path.split(path)[-1].startswith("dummy")
                ),
                os.path.join(temporary_directory, "dummy_file"),
            )



import logging

from octue.mixins import Loggable
from ..base import BaseTestCase


class InheritLoggable(Loggable):
    """Class used purely for testing to check the logger is instantiated"""

    pass


class LoggableTestCase(BaseTestCase):
    def test_instantiates_with_no_args(self):
        """Ensures the class instantiates without arguments"""
        resource = Loggable()
        self.assertIsInstance(resource.logger, logging.Logger)
        self.assertEqual(resource.logger.name, "octue.mixins.loggable")

    def test_inherits_correct_module_name(self):
        """Ensures default logger is the one attached to the file where the derived class was declared,
        not where the Loggable class was declared
        """
        resource = InheritLoggable()
        self.assertIsInstance(resource.logger, logging.Logger)
        self.assertEqual(resource.logger.name, __name__)

    def test_assigns_if_logger_passed(self):
        """Ensures non-default logger is attached correctly"""
        custom_logger = logging.getLogger("custom_logger")
        resource = InheritLoggable(logger=custom_logger)
        self.assertIsInstance(resource.logger, logging.Logger)
        self.assertEqual(resource.logger.name, "custom_logger")



import logging
import os
from datetime import datetime
from google_crc32c import Checksum

from octue.exceptions import FileNotFoundException, InvalidInputException
from octue.mixins import Filterable, Hashable, Identifiable, Loggable, Pathable, Serialisable, Taggable
from octue.utils import isfile
from octue.utils.cloud import storage
from octue.utils.cloud.storage.client import GoogleCloudStorageClient
from octue.utils.cloud.storage.path import generate_gs_path


module_logger = logging.getLogger(__name__)


ID_DEFAULT = None
CLUSTER_DEFAULT = 0
SEQUENCE_DEFAULT = None
TAGS_DEFAULT = None


class Datafile(Taggable, Serialisable, Pathable, Loggable, Identifiable, Hashable, Filterable):
    """Class for representing data files on the Octue system.

    Files in a manifest look like this:

        {
          "path": "folder/subfolder/file_1.csv",
          "cluster": 0,
          "sequence": 0,
          "extension": "csv",
          "tags": "",
          "timestamp": 0,
          "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86",
          "size_bytes": 59684813,
          "sha-512/256": "somesha"
        },

    :parameter float|None timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when
        it was created but could relate to a relevant time point for the data
    :param str id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
    :param logging.Logger logger: A logger instance to which operations with this datafile will be logged. Defaults to
        the module logger.
    :param Union[str, path-like] path: The path of this file, which may include folders or subfolders, within the
        dataset. If no path_from parameter is set, then absolute paths are acceptable, otherwise relative paths are
        required.
    :param Pathable path_from: The root Pathable object (typically a Dataset) that this Datafile's path is relative to.
    :param int cluster: The cluster of files, within a dataset, to which this belongs (default 0)
    :param int sequence: A sequence number of this file within its cluster (if sequences are appropriate)
    :param str tags: Space-separated string of tags relevant to this file
    :param bool skip_checks:

    :return None:
    """

    _ATTRIBUTES_TO_HASH = "name", "cluster", "sequence", "timestamp", "tags"
    _EXCLUDE_SERIALISE_FIELDS = ("logger", "open")

    def __init__(
        self,
        timestamp,
        id=ID_DEFAULT,
        logger=None,
        path=None,
        path_from=None,
        cluster=CLUSTER_DEFAULT,
        sequence=SEQUENCE_DEFAULT,
        tags=TAGS_DEFAULT,
        skip_checks=True,
        **kwargs,
    ):
        super().__init__(
            id=id, hash_value=kwargs.get("hash_value"), logger=logger, tags=tags, path=path, path_from=path_from
        )

        self.cluster = cluster
        self.sequence = sequence
        self.timestamp = timestamp

        if path is None:
            raise InvalidInputException("You must supply a valid 'path' for a Datafile")

        # Set up the file extension or get it from the file path if none passed
        self.extension = self._get_extension_from_path()

        # Run integrity checks on the file
        if not skip_checks:
            self.check(**kwargs)

        self._gcp_metadata = None

    def __lt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.absolute_path < other.absolute_path

    def __gt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.absolute_path > other.absolute_path

    @classmethod
    def from_cloud(cls, project_name, bucket_name, datafile_path, timestamp=None):
        """Instantiate a Datafile from a previously-persisted Datafile in Google Cloud storage. To instantiate a
        Datafile from a regular file on Google Cloud storage, the usage is the same, but include a meaningful value for
        the `timestamp` parameter. The hash for this kind of file will be the Google MD5 hash.

        :param str project_name:
        :param str bucket_name:
        :param str datafile_path: path to file represented by datafile
        :param float|None timestamp:
        :return Datafile:
        """
        metadata = GoogleCloudStorageClient(project_name).get_metadata(bucket_name, datafile_path)
        custom_metadata = metadata.get("metadata") or {}

        datafile = cls(
            timestamp=custom_metadata.get("timestamp", timestamp),
            id=custom_metadata.get("id", ID_DEFAULT),
            path=generate_gs_path(bucket_name, datafile_path),
            hash_value=custom_metadata.get("hash_value", metadata["md5Hash"]),
            cluster=custom_metadata.get("cluster", CLUSTER_DEFAULT),
            sequence=custom_metadata.get("sequence", SEQUENCE_DEFAULT),
            tags=custom_metadata.get("tags", TAGS_DEFAULT),
        )

        datafile._gcp_metadata = metadata
        return datafile

    def to_cloud(self, project_name, bucket_name, path_in_bucket):
        """Upload a datafile to Google Cloud Storage.

        :param str project_name:
        :param str bucket_name:
        :param str path_in_bucket:
        :return str: gs:// path for datafile
        """
        GoogleCloudStorageClient(project_name=project_name).upload_file(
            local_path=self.path,
            bucket_name=bucket_name,
            path_in_bucket=path_in_bucket,
            metadata={
                "timestamp": self.timestamp,
                "id": self.id,
                "hash_value": self.hash_value,
                "cluster": self.cluster,
                "sequence": self.sequence,
                "tags": self.tags.serialise(to_string=False),
            },
        )

        return storage.path.generate_gs_path(bucket_name, path_in_bucket)

    @property
    def name(self):
        return str(os.path.split(self.path)[-1])

    @property
    def _last_modified(self):
        """Get the date/time the file was last modified in units of seconds since epoch (posix time)."""
        if self._path_is_in_google_cloud_storage:
            unparsed_datetime = self._gcp_metadata["updated"]
            parsed_datetime = datetime.strptime(unparsed_datetime, "%Y-%m-%dT%H:%M:%S.%fZ")
            return (parsed_datetime - datetime(1970, 1, 1)).total_seconds()

        return os.path.getmtime(self.absolute_path)

    @property
    def size_bytes(self):
        if self._path_is_in_google_cloud_storage:
            return int(self._gcp_metadata["size"])
        return os.path.getsize(self.absolute_path)

    def __repr__(self):
        return f"<{type(self).__name__}({self.name!r})>"

    def _get_extension_from_path(self, path=None):
        """Gets extension of a file, either from a provided file path or from self.path field"""
        path = path or self.path
        return os.path.splitext(path)[-1].strip(".")

    def _calculate_hash(self):
        """Calculate the hash of the file."""
        hash = Checksum()

        with open(self.absolute_path, "rb") as f:
            # Read and update hash value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                hash.update(byte_block)

        return super()._calculate_hash(hash)

    def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
        """Check file presence and integrity"""
        # TODO Check consistency of size_bytes input against self.size_bytes property for a file if we have one
        # TODO Check consistency of sha against file contents if we have a file
        # TODO Check consistency of last_modified date

        if (extension is not None) and not self.path.endswith(extension):
            raise InvalidInputException(
                f"Extension provided ({extension}) does not match file extension (from {self.path}). Pass extension="
                f"None to set extension from filename automatically."
            )

        if not self.exists():
            raise FileNotFoundException(f"No file found at {self.absolute_path}")

    def exists(self):
        """Returns true if the datafile exists on the current system, false otherwise"""
        return isfile(self.absolute_path)

    @property
    def open(self):
        """Context manager to handle the opening and closing of a Datafile.

        If opened in write mode, the manager will attempt to determine if the folder path exists and, if not, will
        create the folder structure required to write the file.

        Use it like:
        ```
        my_datafile = Datafile(timestamp=None, path='subfolder/subsubfolder/my_datafile.json)
        with my_datafile.open('w') as fp:
            fp.write("{}")
        ```
        This is equivalent to the standard python:
        ```
        my_datafile = Datafile(timestamp=None, path='subfolder/subsubfolder/my_datafile.json)
        os.makedirs(os.path.split(my_datafile.absolute_path)[0], exist_ok=True)
        with open(my_datafile.absolute_path, 'w') as fp:
            fp.write("{}")
        ```
        """

        absolute_path = self.absolute_path

        class DataFileContextManager:
            def __init__(obj, mode="r", **kwargs):
                obj.mode = mode
                obj.kwargs = kwargs
                obj.absolute_path = absolute_path
                if "w" in obj.mode:
                    os.makedirs(os.path.split(obj.absolute_path)[0], exist_ok=True)

            def __enter__(obj):
                obj.fp = open(obj.absolute_path, obj.mode, **obj.kwargs)
                return obj.fp

            def __exit__(obj, *args):
                if obj.fp is not None:
                    obj.fp.close()

        return DataFileContextManager



from twined.utils import TwinedEncoder


class OctueJSONEncoder(TwinedEncoder):
    """A JSON Encoder which allows objects having a `serialise()` method to control their own conversion to primitives"""

    def default(self, obj):

        # If the class defines a serialise() method then use it
        if hasattr(obj, "serialise"):
            return obj.serialise()

        # Otherwise let the base class default method raise the TypeError
        return TwinedEncoder.default(self, obj)
